"""Test the flow accumulator component.

@author: benjaminCampforts
"""
import numpy as np
import pytest
from numpy import testing
from numpy.testing import assert_array_equal

from landlab import FieldError, HexModelGrid, RasterModelGrid
from landlab.components import PriorityFloodFlowRouter
from landlab.components.priority_flood_flow_router.cfuncs import (
    _D8_FlowAcc,
    _D8_flowDir,
)
from landlab.grid.nodestatus import NodeStatus

try:
    PriorityFloodFlowRouter.load_richdem()
except ModuleNotFoundError:
    pytestmark = pytest.mark.skip(reason="richdem is not installed")


def test_check_fields():
    """Check to make sure the right fields have been created."""
    # %%
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    z = mg.add_field(
        "topographic__elevation", mg.node_x**2 + mg.node_y**2, at="node"
    )

    PriorityFloodFlowRouter(mg)
    assert_array_equal(z, mg.at_node["topographic__elevation"])
    assert_array_equal(np.zeros(100), mg.at_node["drainage_area"])
    assert_array_equal(np.ones(100), mg.at_node["water__unit_flux_in"])

    PriorityFloodFlowRouter(mg, runoff_rate=2.0)
    assert_array_equal(np.full(100, 2.0), mg.at_node["water__unit_flux_in"])


# %%
def input_values():
    # %%
    """
    PriorityFloodFlowRouter should throw an error when wrong input values are provided
    """

    # %% Default configuration
    mg1 = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg1.add_field("topographic__elevation", mg1.node_x + mg1.node_y, at="node")

    with pytest.raises(ValueError):
        _ = PriorityFloodFlowRouter(mg1, suppress_out=True, flow_metric="Oops")

    with pytest.raises(ValueError):
        _ = PriorityFloodFlowRouter(mg1, suppress_out=True, depression_handler="Oops")


# %%
def test_seperate_hillflow_update():
    # %%
    """
    Hillflow fields cannot be updated if not initialized
    """

    # %% Default configuration
    mg1 = RasterModelGrid((6, 6), xy_spacing=(1, 1))
    mg1.add_field("topographic__elevation", mg1.node_x + mg1.node_y, at="node")
    fa1 = PriorityFloodFlowRouter(mg1, suppress_out=True)
    fa1.run_one_step()
    with pytest.raises(ValueError):
        fa1.update_hill_fdfa()

    # %%


def test_fields1():
    # %%
    """
    Check to make sure the right fields have been created.

    Check that the sizes are also correct.
    """

    # %% Default configuration
    mg1 = RasterModelGrid((6, 6), xy_spacing=(1, 1))
    mg1.add_field("topographic__elevation", mg1.node_x + mg1.node_y, at="node")
    fa1 = PriorityFloodFlowRouter(mg1, suppress_out=True)
    fa1.run_one_step()
    assert sorted(mg1.at_node) == [
        "depression_free_elevation",
        "drainage_area",
        "flood_status_code",
        "flow__link_to_receiver_node",
        "flow__receiver_node",
        "flow__receiver_proportions",
        "flow__upstream_node_order",
        "squared_length_adjacent",
        "surface_water__discharge",
        "topographic__elevation",
        "topographic__steepest_slope",
        "water__unit_flux_in",
    ]


def test_fields2():
    # %% No flow accumulation
    mg2 = RasterModelGrid((7, 7), xy_spacing=(1, 1))
    mg2.add_field("topographic__elevation", mg2.node_x + mg2.node_y, at="node")
    fa2 = PriorityFloodFlowRouter(mg2, suppress_out=True, accumulate_flow=False)
    fa2.run_one_step()

    assert sorted(mg2.at_node) == [
        "depression_free_elevation",
        "flood_status_code",
        "flow__link_to_receiver_node",
        "flow__receiver_node",
        "flow__receiver_proportions",
        "flow__upstream_node_order",
        "squared_length_adjacent",
        "topographic__elevation",
        "topographic__steepest_slope",
        "water__unit_flux_in",
    ]
    # %%


def test_fields3():
    # %% Second FD (no FA is default)
    mg3 = RasterModelGrid((8, 8), xy_spacing=(1, 1))
    mg3.add_field("topographic__elevation", mg3.node_x + mg3.node_y, at="node")
    fa3 = PriorityFloodFlowRouter(mg3, separate_hill_flow=True, suppress_out=True)
    fa3.run_one_step()
    assert sorted(mg3.at_node) == [
        "depression_free_elevation",
        "drainage_area",
        "flood_status_code",
        "flow__link_to_receiver_node",
        "flow__receiver_node",
        "flow__receiver_proportions",
        "flow__upstream_node_order",
        "hill_flow__receiver_node",
        "hill_flow__receiver_proportions",
        "hill_flow__upstream_node_order",
        "hill_topographic__steepest_slope",
        "squared_length_adjacent",
        "surface_water__discharge",
        "topographic__elevation",
        "topographic__steepest_slope",
        "water__unit_flux_in",
    ]
    # %%


def test_fields4():
    """
    Second FD (with FA )
    """
    mg4 = RasterModelGrid((9, 9), xy_spacing=(1, 1))
    mg4.add_field("topographic__elevation", mg4.node_x + mg4.node_y, at="node")
    fa4 = PriorityFloodFlowRouter(
        mg4, separate_hill_flow=True, accumulate_flow_hill=True, suppress_out=True
    )
    fa4.run_one_step()
    assert sorted(mg4.at_node) == [
        "depression_free_elevation",
        "drainage_area",
        "flood_status_code",
        "flow__link_to_receiver_node",
        "flow__receiver_node",
        "flow__receiver_proportions",
        "flow__upstream_node_order",
        "hill_drainage_area",
        "hill_flow__receiver_node",
        "hill_flow__receiver_proportions",
        "hill_flow__upstream_node_order",
        "hill_surface_water__discharge",
        "hill_topographic__steepest_slope",
        "squared_length_adjacent",
        "surface_water__discharge",
        "topographic__elevation",
        "topographic__steepest_slope",
        "water__unit_flux_in",
    ]


def test_fields5():
    # %% Verify multiple flow
    mg5 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg5.add_field("topographic__elevation", mg5.node_x + mg5.node_y, at="node")
    _ = PriorityFloodFlowRouter(mg5, suppress_out=True, flow_metric="Quinn")

    assert mg5.at_node["topographic__steepest_slope"].shape[1] == 8
    assert mg5.at_node["flow__receiver_node"].shape[1] == 8
    assert mg5.at_node["flow__receiver_proportions"].shape[1] == 8
    assert mg5.at_node["flow__link_to_receiver_node"].shape[1] == 8


def test_accumulated_area_closes():
    """Check that accumulated area is area of core nodes."""
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fa = PriorityFloodFlowRouter(mg, suppress_out=True)
    fa.run_one_step()

    drainage_area = mg.at_node["drainage_area"]
    drained_area = np.sum(drainage_area[mg.boundary_nodes])
    core_area = np.sum(mg.cell_area_at_node[mg.core_nodes])
    assert drained_area == core_area

    # %% multiple flow
    """Check that accumulated area is area of core nodes."""
    mg3 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg3.add_field("topographic__elevation", mg3.node_x + mg3.node_y, at="node")
    fa3 = PriorityFloodFlowRouter(
        mg3,
        separate_hill_flow=True,
        accumulate_flow_hill=True,
        hill_flow_metric="Quinn",
        suppress_out=True,
    )
    fa3.run_one_step()

    drainage_area_hill = mg3.at_node["hill_drainage_area"]
    drainage_area_hill = np.sum(drainage_area_hill[mg3.boundary_nodes])
    core_area = np.sum(mg3.cell_area_at_node[mg3.core_nodes])
    assert drainage_area_hill == core_area


# %%


def test_specifying_routing_method_wrong():
    # %%
    """Test specifying incorrect method for routing compatability"""
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")

    with pytest.raises(TypeError):
        PriorityFloodFlowRouter(mg, flow_director="D2")


# %%
def test_D8_metric():
    # %%
    """Test D8 routing functionality"""

    topographic__elevation = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 21.0, 10.0, 0.0],
            [0.0, 31.0, 30.8, 0.0],
            [0.0, 32.0, 30.9, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )

    mg2 = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    mg2.add_field("topographic__elevation", topographic__elevation, at="node")
    mg2.set_closed_boundaries_at_grid_edges(True, True, True, False)
    fa2 = PriorityFloodFlowRouter(mg2, flow_metric="D8", suppress_out=True)
    fa2.run_one_step()
    pf_r = mg2.at_node["flow__receiver_node"]
    reciever = np.array(
        [
            [0, 1, 2, 3],
            [4, 1, 2, 7],
            [8, 6, 6, 11],
            [12, 14, 10, 15],
            [16, 17, 18, 19],
        ]
    )
    assert_array_equal(reciever.flatten(), pf_r)

    # Compare with default Landlab flow accumulator
    mg1 = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    from landlab.components.flow_accum import FlowAccumulator

    mg1.add_field("topographic__elevation", topographic__elevation, at="node")
    mg1.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa1 = FlowAccumulator(mg1, flow_director="D8")
    fa1.run_one_step()
    defaultLandl_r = mg1.at_node["flow__receiver_node"]
    assert_array_equal(defaultLandl_r, pf_r)


# %%
def test_various_surface_values():
    # %%

    topographic__elevation = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 30.9, 21.0, 0.0],
            [0.0, 30.8, 31.8, 0.0],
            [0.0, 10.0, 11.1, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )

    topography = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 121.0, 110.0, 0.0],
            [0.0, 131.0, 130.8, 0.0],
            [0.0, 132.0, 130.9, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )

    mg2 = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    mg2.add_field("topographic__elevation", topographic__elevation, at="node")
    mg2.add_field("topography", topography, at="node")
    mg2.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa2 = PriorityFloodFlowRouter(
        mg2, flow_metric="D8", suppress_out=True, surface="topography"
    )
    fa2.run_one_step()
    pf_r = mg2.at_node["flow__receiver_node"]
    reciever = np.array(
        [
            [0, 1, 2, 3],
            [4, 1, 2, 7],
            [8, 6, 6, 11],
            [12, 14, 10, 15],
            [16, 17, 18, 19],
        ]
    )
    assert_array_equal(reciever.flatten(), pf_r)

    fa3 = PriorityFloodFlowRouter(
        mg2, flow_metric="D8", suppress_out=True, surface=topography
    )
    fa3.run_one_step()
    pf_r = mg2.at_node["flow__receiver_node"]
    assert_array_equal(reciever.flatten(), pf_r)

    topo_fd = fa3.surface_values
    assert_array_equal(topo_fd, np.reshape(topography, 4 * 5))


# %%
def test_flow_accumulator_properties():
    # %%

    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x * 2 + mg.node_y, at="node")
    fa = PriorityFloodFlowRouter(mg, flow_metric="D8", suppress_out=True)
    # from landlab.components.flow_accum import FlowAccumulator
    # fa = FlowAccumulator(mg)
    fa.run_one_step()

    node_drainage_area = np.array(
        [
            [3.0, 2.0, 1.0, 0.0, 0.0],
            [2.0, 3.0, 2.0, 1.0, 0.0],
            [1.0, 2.0, 2.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    assert_array_equal(fa.node_water_discharge, node_drainage_area.flatten())
    assert_array_equal(fa.node_drainage_area, node_drainage_area.flatten())


def test_flow_accumulator_properties_D4():
    # %% set update_hill_flow_instantaneous to false and update hill flow properties explicitly

    mg = RasterModelGrid((4, 4), xy_spacing=(1, 1))
    _ = mg.add_field("topographic__elevation", mg.node_x * 2 + mg.node_y, at="node")
    fa = PriorityFloodFlowRouter(
        mg,
        flow_metric="D4",
        suppress_out=True,
        separate_hill_flow=True,
        hill_flow_metric="D4",
        accumulate_flow_hill=True,
        update_hill_flow_instantaneous=False,
    )
    # from landlab.components.flow_accum import FlowAccumulator
    # fa = FlowAccumulator(mg)
    fa.run_one_step()
    fa.update_hill_fdfa()

    node_drainage_area = np.array(
        [0.0, 0.0, 0.0, 0.0, 2.0, 2.0, 1.0, 0.0, 2.0, 2.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )

    assert_array_equal(fa.node_water_discharge, node_drainage_area.flatten())
    assert_array_equal(fa.node_drainage_area, node_drainage_area.flatten())

    # Hill flow accumulation should be similar
    assert_array_equal(
        mg.at_node["hill_surface_water__discharge"], node_drainage_area.flatten()
    )
    assert_array_equal(mg.at_node["hill_drainage_area"], node_drainage_area.flatten())


# %%
def test_flow_accumulator_properties_D4_varying_runoff():
    # %%

    mg = RasterModelGrid((4, 4), xy_spacing=(1, 1))
    _ = mg.add_field("topographic__elevation", mg.node_x * 2 + mg.node_y, at="node")
    runoff_rate = mg.node_x * 0 + 2
    fa = PriorityFloodFlowRouter(
        mg,
        flow_metric="D4",
        suppress_out=True,
        separate_hill_flow=True,
        hill_flow_metric="D4",
        runoff_rate=runoff_rate,
        accumulate_flow_hill=True,
    )
    # from landlab.components.flow_accum import FlowAccumulator
    # fa = FlowAccumulator(mg)
    fa.run_one_step()

    node_drainage_area = np.array(
        [0.0, 0.0, 0.0, 0.0, 2.0, 2.0, 1.0, 0.0, 2.0, 2.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )

    assert_array_equal(fa.node_water_discharge, 2 * node_drainage_area.flatten())
    assert_array_equal(fa.node_drainage_area, node_drainage_area.flatten())

    # Hill flow accumulation should be similar
    assert_array_equal(
        mg.at_node["hill_surface_water__discharge"], 2 * node_drainage_area.flatten()
    )
    assert_array_equal(mg.at_node["hill_drainage_area"], node_drainage_area.flatten())


def test_flow_accumulator_properties_D4_varying_water__unit_flux_in():
    # %%

    mg = RasterModelGrid((4, 4), xy_spacing=(1, 1))
    _ = mg.add_field("topographic__elevation", mg.node_x * 2 + mg.node_y, at="node")
    mg.add_field("water__unit_flux_in", mg.node_x * 0 + 2, at="node")
    fa = PriorityFloodFlowRouter(
        mg,
        flow_metric="D4",
        suppress_out=True,
        separate_hill_flow=True,
        hill_flow_metric="D4",
        accumulate_flow_hill=True,
    )
    # from landlab.components.flow_accum import FlowAccumulator
    # fa = FlowAccumulator(mg)
    fa.run_one_step()

    node_drainage_area = np.array(
        [0.0, 0.0, 0.0, 0.0, 2.0, 2.0, 1.0, 0.0, 2.0, 2.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )

    assert_array_equal(fa.node_water_discharge, 2 * node_drainage_area.flatten())
    assert_array_equal(fa.node_drainage_area, node_drainage_area.flatten())

    # Hill flow accumulation should be similar
    assert_array_equal(
        mg.at_node["hill_surface_water__discharge"], 2 * node_drainage_area.flatten()
    )
    assert_array_equal(mg.at_node["hill_drainage_area"], node_drainage_area.flatten())


# %%
def test_flow_accumulator_varying_Runoff_rate():
    """
    There are two options to provide a runoff value
    The first one is to provide a weight array as a runoff_value input argument
    """

    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x * 2 + mg.node_y, at="node")
    runoff_rate = mg.node_x * 0 + 2
    fa = PriorityFloodFlowRouter(
        mg,
        flow_metric="D8",
        suppress_out=True,
        runoff_rate=runoff_rate,
        separate_hill_flow=True,
        hill_flow_metric="D8",
        accumulate_flow_hill=True,
    )

    # from landlab.components.flow_accum import FlowAccumulator
    # fa = FlowAccumulator(mg)
    fa.run_one_step()

    node_drainage_area = np.array(
        [
            [3.0, 2.0, 1.0, 0.0, 0.0],
            [2.0, 3.0, 2.0, 1.0, 0.0],
            [1.0, 2.0, 2.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    assert_array_equal(fa.node_water_discharge, 2 * node_drainage_area.flatten())
    assert_array_equal(fa.node_drainage_area, node_drainage_area.flatten())

    # Hill flow accumulation should be similar
    assert_array_equal(
        mg.at_node["hill_surface_water__discharge"], 2 * node_drainage_area.flatten()
    )
    assert_array_equal(mg.at_node["hill_drainage_area"], node_drainage_area.flatten())


# %%
def test_flow_accumulator_varying_water__unit_flux_in():
    """
    There are two options to provide a runoff value
    The second is to provide a weight array as a water__unit_flux_in input field
    """

    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x * 2 + mg.node_y, at="node")
    mg.add_field("water__unit_flux_in", mg.node_x * 0 + 2, at="node")
    fa = PriorityFloodFlowRouter(
        mg,
        flow_metric="D8",
        suppress_out=True,
        separate_hill_flow=True,
        hill_flow_metric="D8",
        accumulate_flow_hill=True,
    )

    # from landlab.components.flow_accum import FlowAccumulator
    # fa = FlowAccumulator(mg)
    fa.run_one_step()

    node_drainage_area = np.array(
        [
            [3.0, 2.0, 1.0, 0.0, 0.0],
            [2.0, 3.0, 2.0, 1.0, 0.0],
            [1.0, 2.0, 2.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    assert_array_equal(fa.node_water_discharge, 2 * node_drainage_area.flatten())
    assert_array_equal(fa.node_drainage_area, node_drainage_area.flatten())

    # Hill flow accumulation should be similar
    assert_array_equal(
        mg.at_node["hill_surface_water__discharge"], 2 * node_drainage_area.flatten()
    )
    assert_array_equal(mg.at_node["hill_drainage_area"], node_drainage_area.flatten())


# %%
def test_flow_accumulator_properties_breach():
    # %%

    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x * 2 + mg.node_y, at="node")
    fa = PriorityFloodFlowRouter(
        mg, flow_metric="D8", suppress_out=True, depression_handler="breach"
    )
    # from landlab.components.flow_accum import FlowAccumulator
    # fa = FlowAccumulator(mg)
    fa.run_one_step()

    node_drainage_area = np.array(
        [
            [3.0, 2.0, 1.0, 0.0, 0.0],
            [2.0, 3.0, 2.0, 1.0, 0.0],
            [1.0, 2.0, 2.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    assert_array_equal(fa.node_water_discharge, node_drainage_area.flatten())
    assert_array_equal(fa.node_drainage_area, node_drainage_area.flatten())


# %%
def test_bad_metric_name():
    # %%
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(ValueError):
        PriorityFloodFlowRouter(mg, flow_metric="spam", suppress_out=True)


# %%
def test_bad_hill_metric_name():
    # %%
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(ValueError):
        PriorityFloodFlowRouter(
            mg, hill_flow_metric="Landlab_is_cool", suppress_out=True
        )


# %%
def test_bad_depression_handler():
    # %%
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(ValueError):
        PriorityFloodFlowRouter(
            mg, depression_handler="depression_filler", suppress_out=True
        )


# %%
def test_hex_mfd():
    # %%
    mg = HexModelGrid((5, 3))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(FieldError):
        PriorityFloodFlowRouter(mg, flow_metric="MFD", suppress_out=True)
    # %%


def test_sum_prop_is_one():
    # %%
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fa = PriorityFloodFlowRouter(mg, separate_hill_flow=True, suppress_out=True)
    fa.run_one_step()

    # Single flow
    props_Pf = mg.at_node["flow__receiver_proportions"]
    testing.assert_array_almost_equal(
        props_Pf,
        np.ones_like(props_Pf),
        decimal=5,
        err_msg="Sum of flow proportions is not equal to one",
        verbose=True,
    )

    # Multiple flow
    props_Pf = mg.at_node["hill_flow__receiver_proportions"]
    props_Pf[props_Pf == -1] = 0
    props_Pf = props_Pf.sum(axis=1)
    testing.assert_array_almost_equal(
        props_Pf,
        np.ones_like(props_Pf),
        decimal=5,
        err_msg="Sum of flow proportions is not equal to one",
        verbose=True,
    )

    # %% multiple flow with D8 over hills
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fa = PriorityFloodFlowRouter(
        mg, separate_hill_flow=True, hill_flow_metric="D8", suppress_out=True
    )
    fa.run_one_step()

    # Multiple flow
    props_Pf = mg.at_node["hill_flow__receiver_proportions"]
    testing.assert_array_almost_equal(
        props_Pf,
        np.ones_like(props_Pf),
        decimal=5,
        err_msg="Sum of flow proportions is not equal to one",
        verbose=True,
    )

    # %% multiple flow with D8 over rivers
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fa = PriorityFloodFlowRouter(mg, flow_metric="Quinn", suppress_out=True)
    fa.run_one_step()

    # Multiple flow
    props_Pf = mg.at_node["flow__receiver_proportions"]
    props_Pf[props_Pf == -1] = 0
    props_Pf = np.sum(props_Pf, axis=1)
    testing.assert_array_almost_equal(
        props_Pf,
        np.ones_like(props_Pf),
        decimal=5,
        err_msg="Sum of flow proportions is not equal to one",
        verbose=True,
    )
    # %%


def test_cython_functions():
    # %% test d8 flow dir with open boundaries
    mg = RasterModelGrid((4, 4), xy_spacing=(1, 1))
    z = mg.add_zeros("topographic__elevation", at="node")
    z[5:7] = [1, 3]
    z[9:11] = [3, 4]
    from landlab.plot import imshow_grid

    imshow_grid(mg, "topographic__elevation")

    activeCells = np.array(mg.status_at_node != NodeStatus.CLOSED + 0, dtype=int)
    receivers = np.array(mg.status_at_node, dtype=int)
    distance_receiver = np.zeros((receivers.shape), dtype=float)
    cores = mg.core_nodes
    activeCores = cores[activeCells[cores] == 1]
    # Make boundaries to save time with conditionals in c loops
    receivers[np.nonzero(mg.status_at_node)] = -1
    steepest_slope = np.zeros((receivers.shape), dtype=float)
    el_dep_free = z
    el_ori = mg.at_node["topographic__elevation"]
    dist = (
        np.array([1, np.sqrt(2), 1, np.sqrt(2), 1, np.sqrt(2), 1, np.sqrt(2)]) * mg.dx
    )

    ngb = np.zeros((8,), dtype=int)
    el_d = np.zeros((8,), dtype=float)

    # Links
    adj_link = np.array(mg.d8s_at_node, dtype=int)
    recvr_link = np.zeros((receivers.shape), dtype=int) - 1

    _D8_flowDir(
        receivers,
        distance_receiver,
        steepest_slope,
        np.array(el_dep_free),
        el_ori,
        dist,
        ngb,
        activeCores,
        activeCells,
        el_d,
        mg.number_of_node_columns,
        mg.dx,
        adj_link,
        recvr_link,
    )

    # We know where the water will flow using the D8 steepest descent algo
    # Also consider that under equal slopes, the flow will follow Landlab's
    # rotational ordering going first to cardial, then to diagonal cells
    known_rec = np.array([-1, -1, -1, -1, -1, 4, 7, -1, -1, 13, 11, -1, -1, -1, -1, -1])

    testing.assert_array_equal(
        known_rec,
        receivers,
        err_msg="Error with D8 flow routing calculations",
        verbose=True,
    )

    # %% test flow acc with open boundaries
    node_cell_area = np.array(mg.cell_area_at_node)
    node_cell_area[mg.closed_boundary_nodes] = 0.0
    dis = np.full(mg.number_of_nodes, node_cell_area)

    da = np.array(node_cell_area)
    sort = np.argsort(el_dep_free)
    stack_flip = np.flip(sort)
    # Filter out donors giving to receivers being -1
    stack_flip = np.array(stack_flip[receivers[stack_flip] != -1], dtype=int)

    _D8_FlowAcc(da, dis, stack_flip, receivers)

    # We know how much water will accumualte given the calculated flow dir (see before)
    known_FA = np.array(
        [0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 1.0, 1.0, 1.0, 0.0, 1.0, 0.0, 0.0]
    )

    testing.assert_array_equal(
        known_FA,
        da,
        err_msg="Error with D8_FlowAcc calculations",
        verbose=True,
    )

    # %% test d8 flow dir with closed boundaries
    mg = RasterModelGrid((4, 4), xy_spacing=(1, 1))
    z = mg.add_zeros("topographic__elevation", at="node")
    # Close all model boundary edges
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    # Set lower-left (southwest) corner as an open boundary
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )

    z[5:7] = [1, 3]
    z[9:11] = [3, 4]

    activeCells = np.array(mg.status_at_node != NodeStatus.CLOSED + 0, dtype=int)
    receivers = np.array(mg.status_at_node, dtype=int)
    distance_receiver = np.zeros((receivers.shape), dtype=float)
    cores = mg.core_nodes
    activeCores = cores[activeCells[cores] == 1]
    # Make boundaries to save time with conditionals in c loops
    receivers[np.nonzero(mg.status_at_node)] = -1
    steepest_slope = np.zeros((receivers.shape), dtype=float)
    el_dep_free = z
    el_ori = mg.at_node["topographic__elevation"]
    dist = (
        np.array([1, np.sqrt(2), 1, np.sqrt(2), 1, np.sqrt(2), 1, np.sqrt(2)]) * mg.dx
    )

    ngb = np.zeros((8,), dtype=int)
    el_d = np.zeros((8,), dtype=float)

    # Links
    adj_link = np.array(mg.d8s_at_node, dtype=int)
    recvr_link = np.zeros((receivers.shape), dtype=int) - 1

    _D8_flowDir(
        receivers,
        distance_receiver,
        steepest_slope,
        np.array(el_dep_free),
        el_ori,
        dist,
        ngb,
        activeCores,
        activeCells,
        el_d,
        mg.number_of_node_columns,
        mg.dx,
        adj_link,
        recvr_link,
    )

    # We know where the water will flow using the D8 steepest descent algo
    # Also consider that under equal slopes, the flow will follow Landlab's
    # rotational ordering going first to cardial, then to diagonal cells
    known_rec = np.array([-1, -1, -1, -1, -1, 0, 5, -1, -1, 5, 5, -1, -1, -1, -1, -1])

    testing.assert_array_equal(
        known_rec,
        receivers,
        err_msg="Error with D8 flow routing calculations",
        verbose=True,
    )

    # %% test flow acc with closed boundaries
    node_cell_area = np.array(mg.cell_area_at_node)
    node_cell_area[mg.closed_boundary_nodes] = 0.0
    dis = np.full(mg.number_of_nodes, node_cell_area)

    da = np.array(node_cell_area)
    sort = np.argsort(el_dep_free)
    stack_flip = np.flip(sort)
    # Filter out donors giving to receivers being -1
    stack_flip = np.array(stack_flip[receivers[stack_flip] != -1], dtype=int)

    _D8_FlowAcc(da, dis, stack_flip, receivers)

    # We know how much water will accumualte given the calculated flow dir (see before)
    known_FA = np.array(
        [4.0, 0.0, 0.0, 0.0, 0.0, 4.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )

    testing.assert_array_equal(
        known_FA,
        da,
        err_msg="Error with D8_FlowAcc calculations",
        verbose=True,
    )



from landlab.graph.framed_voronoi.framed_voronoi import HorizontalRectVoronoiGraph


def pytest_generate_tests(metafunc):
    if "layout_graph" in metafunc.fixturenames:
        metafunc.parametrize(
            "layout_graph",
            [HorizontalRectVoronoiGraph],
        )



import numpy as np
from scipy.sparse.linalg import spsolve

from landlab import Component, HexModelGrid
from landlab.grid.diagonals import DiagonalsMixIn


def make_empty_matrix_and_rhs(grid):
    from scipy.sparse import csc_matrix

    mat = csc_matrix((grid.number_of_core_nodes, grid.number_of_core_nodes))
    rhs = np.zeros(grid.number_of_core_nodes)
    return mat, rhs


def zero_out_matrix(grid, mat, rcvr, mat_id):
    for i in grid.core_nodes:
        j = mat_id[i]
        mat[j, j] = 0.0
        r = rcvr[i]
        if grid.status_at_node[r] == grid.BC_NODE_IS_CORE:
            k = mat_id[r]
            mat[j, k] = 0.0
            mat[k, k] = 0.0
            mat[k, j] = 0.0


class GravelRiverTransporter(Component):
    """Model drainage network evolution for a network of transport-limited
    gravel-bed rivers with downstream abrasion.

    GravelRiverTransporter is designed to operate together with a flow-routing
    component such as PriorityFloodFlowRouter, so that each grid node has
    a defined flow direction toward one of its neighbor nodes. Each core node
    is assumed to contain one outgoing fluvial channel, and (depending on
    the drainage structure) zero, one, or more incoming channels. These channels are
    treated as effectively sub-grid-scale features that are embedded in valleys
    that have a width of one grid cell. The rate of gravel transport out of
    a given node is calculated as the product of bankfull discharge, channel
    gradient (to the 7/6 power), a dimensionless transport coefficient, and
    an intermittency factor that represents the fraction of time that bankfull
    flow occurs. The derivation of the transport law is given by Wickert &
    Schildgen (2019), and it derives from the assumption that channels are
    gravel-bedded and that they "instantaneously" adjust their width such that
    bankfull bed shear stress is just slightly higher than the threshold for
    grain motion. The substrate is assumed to consist entirely of gravel-size
    material with a given bulk porosity. The component calculates the loss of
    gravel-sized material to abrasion (i.e., conversion to finer sediment, which
    is not explicitly tracked) as a function of the volumetric transport rate,
    an abrasion coefficient with units of inverse length, and the local transport
    distance (for example, if a grid node is carrying a gravel load Qs to a
    neighboring node dx meters downstream, the rate of gravel loss in volume per
    time per area at the node will be beta Qs dx, where beta is the abrasion
    coefficient). Sediment mass conservation is calculated across each entire
    grid cell. For example, if a cell has surface area A, a total volume influx
    Qin, and downstream transport rate Qs, the resulting rate of change of
    elevation will be (Qin - Qs / (A (1 - phi)), where phi is porosity.

    Parameters
    ----------
    grid : ModelGrid
        A Landlab model grid object
    intermittency_factor : float (default 0.01)
        Fraction of time that bankfull flow occurs
    transport_coefficient : float (default 0.041)
        Dimensionless transport efficiency factor (see Wickert & Schildgen 2019)
    abrasion_coefficient : float (default 0.0 1/m)
        Abrasion coefficient with units of inverse length
    sediment_porosity : float (default 0.35)
        Bulk porosity of bed sediment
    solver : string (default "explicit")
        Solver type (currently only "explicit" is tested and operational)

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator
    >>> grid = RasterModelGrid((3, 3), xy_spacing=1000.0)
    >>> elev = grid.add_zeros("topographic__elevation", at="node")
    >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
    >>> grid.status_at_node[5] = grid.BC_NODE_IS_FIXED_VALUE
    >>> fa = FlowAccumulator(grid, runoff_rate=10.0)
    >>> fa.run_one_step()
    >>> transporter = GravelRiverTransporter(grid, abrasion_coefficient=0.0005)
    >>> for _ in range(200):
    ...     fa.run_one_step()
    ...     elev[grid.core_nodes] += 1.0
    ...     transporter.run_one_step(10000.0)
    >>> int(elev[4] * 100)
    2366
    """

    _ONE_SIXTH = 1.0 / 6.0

    _name = "GravelRiverTransporter"

    _unit_agnostic = True

    _info = {
        "bedload_sediment__rate_of_loss_to_abrasion": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/y",
            "mapping": "node",
            "doc": "Rate of bedload sediment volume loss to abrasion per unit area",
        },
        "bedload_sediment__volume_influx": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**2/y",
            "mapping": "node",
            "doc": "Volumetric incoming streamwise bedload sediment transport rate",
        },
        "bedload_sediment__volume_outflux": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**2/y",
            "mapping": "node",
            "doc": "Volumetric outgoing streamwise bedload sediment transport rate",
        },
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "sediment__rate_of_change": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/y",
            "mapping": "node",
            "doc": "Time rate of change of sediment thickness",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**3/y",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
    }

    def __init__(
        self,
        grid,
        intermittency_factor=0.01,
        transport_coefficient=0.041,
        abrasion_coefficient=0.0,
        sediment_porosity=0.35,
        solver="explicit",
    ):
        """Initialize GravelRiverTransporter."""

        super().__init__(grid)

        # Parameters
        self._trans_coef = transport_coefficient
        self._intermittency_factor = intermittency_factor
        self._abrasion_coef = abrasion_coefficient
        self._porosity_factor = 1.0 / (1.0 - sediment_porosity)

        # Fields and arrays
        self._elev = grid.at_node["topographic__elevation"]
        self._discharge = grid.at_node["surface_water__discharge"]
        self._slope = grid.at_node["topographic__steepest_slope"]
        self._receiver_node = grid.at_node["flow__receiver_node"]
        self._receiver_link = grid.at_node["flow__link_to_receiver_node"]
        super().initialize_output_fields()
        self._sediment_influx = grid.at_node["bedload_sediment__volume_influx"]
        self._sediment_outflux = grid.at_node["bedload_sediment__volume_outflux"]
        self._dzdt = grid.at_node["sediment__rate_of_change"]
        self._abrasion = grid.at_node["bedload_sediment__rate_of_loss_to_abrasion"]

        # Constants
        self._SEVEN_SIXTHS = 7.0 / 6.0

        # Solver type
        if solver == "explicit":
            self.run_one_step = self.run_one_step_simple_explicit
        elif solver == "matrix":
            import warnings

            from landlab.utils.matrix import get_core_node_at_node

            warnings.warn(
                "Matrix-based solver is experimental & not fully tested", stacklevel=2
            )
            self.run_one_step = self.run_one_step_matrix_inversion
            self._mat, self._rhs = make_empty_matrix_and_rhs(grid)
            self._mat_id = np.zeros(grid.number_of_nodes, dtype=int)
            self._mat_id = get_core_node_at_node(grid)
        else:
            raise ValueError("Solver type not recognized")

        self._setup_length_of_flow_link()

    def _setup_length_of_flow_link(self):
        """Set up a float or array containing length of the flow link from each node,
        which is needed for the abrasion rate calculations.
        """
        if isinstance(self.grid, HexModelGrid):
            self._flow_link_length_over_cell_area = (
                self.grid.spacing / self.grid.area_of_cell[0]
            )
            self._flow_length_is_variable = False
        elif isinstance(self.grid, DiagonalsMixIn):
            self._flow_length_is_variable = True
            self._grid_has_diagonals = True
        else:
            self._flow_length_is_variable = True
            self._grid_has_diagonals = False

    def _update_flow_link_length_over_cell_area(self):
        """Update the ratio of the length of link along which water flows out of
        each node to the area of the node's cell."""
        if self._grid_has_diagonals:
            self._flow_link_length_over_cell_area = (
                self.grid.length_of_d8[self._receiver_link[self.grid.core_nodes]]
                / self.grid.area_of_cell[self.grid.cell_at_node[self.grid.core_nodes]]
            )
        else:
            self._flow_link_length_over_cell_area = (
                self.grid.length_of_link[self._receiver_link[self.grid.core_nodes]]
                / self.grid.area_of_cell[self.grid.cell_at_node[self.grid.core_nodes]]
            )

    def calc_implied_depth(self, grain_diameter=0.01):
        """Utility function that calculates and returns water depth implied by
        slope and grain diameter, using Wickert & Schildgen (2019) equation 8.

        The equation is

            h = ((rho_s - rho / rho)) (1 + epsilon) tau_c* (D / S)

        where the factors on the right are sediment and water density, excess
        shear-stress factor, critical Shields stress, grain diameter, and slope
        gradient. Here the prefactor on D/S assumes sediment density of 2650 kg/m3,
        water density of 1000 kg/m3, shear-stress factor of 0.2, and critical
        Shields stress of 0.0495, giving a value of 0.09801.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 3), xy_spacing=1000.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[3:] = 10.0
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = GravelRiverTransporter(grid)
        >>> depth = transporter.calc_implied_depth()
        >>> int(depth[4] * 1000)
        98
        """
        DEPTH_FACTOR = 0.09801
        depth = np.zeros(self._grid.number_of_nodes)
        nonzero_slope = self._slope > 0.0
        depth[nonzero_slope] = (
            DEPTH_FACTOR * grain_diameter / self._slope[nonzero_slope]
        )
        return depth

    def calc_implied_width(self, grain_diameter=0.01, time_unit="y"):
        """Utility function that calculates and returns channel width implied by
        discharge, slope, and grain diameter, using Wickert & Schildgen (2019)
        equation 16.

        The equation is

            b = kb Q S**(7/6) / D**(3/2)

        where the dimensional prefactor, which includes sediment and water
        density, gravitational acceleration, critical Shields stress, and the
        transport factor epsilon, is

            kb = 0.17 g**(-1/2) (((rho_s - rho) / rho) (1 + eps) tau_c*)**(-5/3)

        Using g = 9.8 m/s2, rho_s = 2650 (quartz), rho = 1000 kg/m3, eps = 0.2,
        and tau_c* = 0.0495, kb ~ 2.61 s/m**(1/2). Converting to years,
        kb = 8.26e-8.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 3), xy_spacing=10000.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[3:] = 100.0
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = GravelRiverTransporter(grid)
        >>> width = transporter.calc_implied_width()
        >>> int(width[4] * 100)
        3833
        >>> grid.at_node["surface_water__discharge"] *= 1. / (3600 * 24 * 365.25)
        >>> width = transporter.calc_implied_width(time_unit="s")
        >>> int(width[4] * 100)
        3838
        """
        if time_unit[0] == "y":
            width_fac = 8.26e-8
        else:
            width_fac = 2.61  # assume seconds if not years
        width = np.zeros(self._grid.number_of_nodes)
        width = (
            width_fac
            * self._discharge
            * self._slope ** (7.0 / 6.0)
            / (grain_diameter**1.5)
        )
        return width

    def calc_transport_capacity(self):
        """Calculate and return bed-load transport capacity.

        Calculation uses Wickert-Schildgen approach, and provides
        volume per time rate.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 3), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[3:] = 1.0
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = GravelRiverTransporter(grid)
        >>> transporter.calc_transport_capacity()
        >>> round(transporter._sediment_outflux[4], 4)
        0.019
        """
        self._sediment_outflux[:] = (
            self._trans_coef
            * self._intermittency_factor
            * self._discharge
            * self._slope**self._SEVEN_SIXTHS
        )

    def calc_abrasion_rate(self):
        """Update the rate of bedload loss to abrasion, per unit area.

        Here we use the average of incoming and outgoing sediment flux to
        calculate the loss rate to abrasion.

        The factor dx (node spacing) appears in the denominator to represent
        flow segment length (i.e., length of the link along which water is
        flowing in the cell) divided by cell area. This would need to be updated
        to handle non-raster and/or non-uniform grids.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 3), xy_spacing=1000.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[3:] = 10.0
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = GravelRiverTransporter(grid, abrasion_coefficient=0.0002)
        >>> transporter.calc_transport_capacity()
        >>> transporter.calc_abrasion_rate()
        >>> int(transporter._abrasion[4] * 1e8)
        19
        """
        cores = self._grid.core_nodes
        if self._flow_length_is_variable:
            self._update_flow_link_length_over_cell_area()
        self._abrasion[cores] = (
            self._abrasion_coef
            * 0.5
            * (self._sediment_outflux[cores] + self._sediment_influx[cores])
            * self._flow_link_length_over_cell_area
        )

    def calc_sediment_rate_of_change(self):
        """Update the rate of thickness change of coarse sediment at each core node.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 4), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[:] = 0.01 * grid.x_of_node
        >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
        >>> grid.status_at_node[4] = grid.BC_NODE_IS_FIXED_VALUE
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = GravelRiverTransporter(grid)
        >>> transporter.calc_sediment_rate_of_change()
        >>> np.round(transporter._sediment_outflux[4:7], 3)
        array([ 0.   ,  0.038,  0.019])
        >>> np.round(transporter._sediment_influx[4:7], 3)
        array([ 0.038,  0.019,  0.   ])
        >>> np.round(transporter._dzdt[5:7], 8)
        array([ -2.93000000e-06,  -2.93000000e-06])
        """
        self.calc_transport_capacity()
        if self._abrasion_coef > 0.0:
            self.calc_abrasion_rate()
        cores = self.grid.core_nodes
        self._sediment_influx[:] = 0.0
        for c in cores:  # send sediment downstream
            r = self._receiver_node[c]
            self._sediment_influx[r] += self._sediment_outflux[c]
        self._dzdt[cores] = self._porosity_factor * (
            (self._sediment_influx[cores] - self._sediment_outflux[cores])
            / self.grid.area_of_cell[self.grid.cell_at_node[cores]]
            - self._abrasion[cores]
        )

    def run_one_step_simple_explicit(self, dt):
        """Advance solution by time interval dt.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 4), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[:] = 0.01 * grid.x_of_node
        >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
        >>> grid.status_at_node[4] = grid.BC_NODE_IS_FIXED_VALUE
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = GravelRiverTransporter(grid, solver="explicit")
        >>> transporter.run_one_step(1000.0)
        >>> np.round(elev[4:7], 4)
        array([ 0.    ,  0.9971,  1.9971])
        """
        self.calc_sediment_rate_of_change()
        self._elev += self._dzdt * dt

    def _fill_matrix_and_rhs(self, dt):
        """Fill out entries in a sparse matrix and corresponding right-hand side
        vector.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 4), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[:] = 0.01 * grid.x_of_node
        >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
        >>> grid.status_at_node[4] = grid.BC_NODE_IS_FIXED_VALUE
        >>> fa = FlowAccumulator(grid)
        >>> transporter = GravelRiverTransporter(grid, solver="matrix")
        >>> transporter._mat.toarray()
        array([[ 0.,  0.],
               [ 0.,  0.]])
        >>> fa.run_one_step()
        >>> transporter._receiver_node[5:7]
        array([4, 5])
        >>> transporter._fill_matrix_and_rhs(1000.0)
        >>> transporter._rhs
        array([ 1.,  2.])
        """
        prefac = (
            self._trans_coef * self._intermittency_factor * self._porosity_factor * dt
        ) / self.grid.dx**2
        a = prefac * (1.0 / self.grid.dx + self._abrasion_coef / 2)
        b = prefac * (1.0 / self.grid.dx - self._abrasion_coef / 2)
        f = self._discharge * (self._slope ** (self._ONE_SIXTH))

        zero_out_matrix(self.grid, self._mat, self._receiver_node, self._mat_id)
        for i in self.grid.core_nodes:
            j = self._mat_id[i]
            self._rhs[j] = self._elev[i]
            self._mat[j, j] += 1 + a * f[i]
            r = self._receiver_node[i]
            if self.grid.status_at_node[r] == self.grid.BC_NODE_IS_CORE:
                k = self._mat_id[r]
                self._mat[j, k] -= a * f[i]
                self._mat[k, k] += b * f[i]
                self._mat[k, j] -= b * f[i]
            else:
                self._rhs[j] += a * f[i] * self._elev[r]

    def run_one_step_matrix_inversion(self, dt):
        """Advance solution by time interval dt.

        WARNING: EXPERIMENTAL AND NOT FULLY TESTED - USE AT OWN RISK!

        Notes
        -----
        Does not update abrasion rate or sediment outflux fields.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> grid = RasterModelGrid((3, 4), xy_spacing=100.0)
        >>> elev = grid.add_zeros("topographic__elevation", at="node")
        >>> elev[:] = 0.01 * grid.x_of_node
        >>> grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
        >>> grid.status_at_node[4] = grid.BC_NODE_IS_FIXED_VALUE
        >>> fa = FlowAccumulator(grid)
        >>> fa.run_one_step()
        >>> transporter = GravelRiverTransporter(grid, solver="matrix")
        >>> transporter.run_one_step == transporter.run_one_step_matrix_inversion
        True
        >>> transporter.run_one_step(1000.0)
        """
        self._fill_matrix_and_rhs(dt)
        self._elev[self.grid.core_nodes] = spsolve(self._mat, self._rhs)



import pathlib

import numpy as np
import pytest

from landlab import VoronoiDelaunayGrid


def test_save(tmpdir):
    with tmpdir.as_cwd():
        grid = VoronoiDelaunayGrid(np.random.rand(20), np.random.rand(20))
        filename = grid.save("mytestsave.grid")

        assert pathlib.Path(filename).is_file()


def test_save_with_clobber(tmpdir):
    with tmpdir.as_cwd():
        pathlib.Path("mytestsave.grid").touch()

        grid = VoronoiDelaunayGrid(np.random.rand(20), np.random.rand(20))
        with pytest.raises(ValueError):
            grid.save("mytestsave.grid")
        filename = grid.save("mytestsave.grid", clobber=True)

        assert pathlib.Path(filename).is_file()
        assert pathlib.Path(filename).stat().st_size > 0


def test_save_with_extension(tmpdir):
    with tmpdir.as_cwd():
        grid = VoronoiDelaunayGrid(np.random.rand(20), np.random.rand(20))
        filename = grid.save("mytestsave", clobber=True)

        assert filename == "mytestsave.grid"
        assert pathlib.Path(filename).is_file()

        filename = grid.save("mytestsave.foo", clobber=True)
        assert filename == "mytestsave.foo.grid"
        assert pathlib.Path(filename).is_file()



#distutils: define_macros=NPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION
#distutils: extra_compile_args = -std=c++11
#distutils: extra_link_args = -std=c++11
""" Contains the cython functions for the component method
Flow_router.run_directions(). Flow directions and depression handling are done
adapting Barnes et al., 2014 algorithm #4."""

import numpy as np
cimport numpy as cnp
cimport cython
from libcpp cimport bool
from libcpp.pair cimport pair

# 1-3. Instantiate the Queues (Steps noted #1-3, #5-6 in Barnes, 2014's algorithm #4
####################################################################################
cdef extern from "_priority_queue.hpp" nogil:
    cdef cppclass _priority_queue:
        _priority_queue(...) except +
        void push(pair[cnp.int_t, cnp.float_t])
        pair[cnp.int_t, cnp.float_t] top() except +
        void pop()
        bool empty()
        cnp.int_t size()

cdef bool _compare_second(pair[int, double] a, pair[int, double] b) nogil:
    return a.second > b.second


@cython.boundscheck(False) # turn off bounds-checking for entire function
cdef void _init_flow_direction_queues(
    const cnp.int_t [:] base_level_nodes, const cnp.int_t [:] closed_nodes,
    cnp.float_t [:] z, _priority_queue& to_do, cnp.int_t [:] receivers,
    cnp.int_t [:] outlet_nodes, cnp.int_t [:] done, cnp.int_t* done_n_ptr) nogil:
    """
    Add the base-level nodes to the queue and update receivers for base-level and
    closed nodes. Updates to_do, receivers, outlet_nodes, done and the value pointed
    by done_n_ptr.

    Implementation remarks: It's important to pass the priority_queue to_do as a
    reference, otherwise it won't be modified. I didn't manage to use the bint type
    for the variable done, so here 0 is for False and 1 for True.

    Params
    ------
    base_level_nodes: memoryview(long).
        Base-level nodes.
    closed_nodes: memoryview(long).
        Closed nodes.
    z: memoryview(double).
        Elevation of eachn
    to_do: _priority_queue&
        Queue which handles the nodes where to direct flow
    receivers: memoryview(long).
        Receiver of each node (ordered by node id).
    outlet_nodes: memoryview(long).
        Base-level outlet of each node (ordered by node id).
    done: memoryview(long).
        Flag for each node (ordered by node id). If 1, the node has been done,
        otherwise 0.
    done_n_ptr: long*
        Pointer to the number of done nodes. Necessary for multithreading
        (future evolution).
    """

    cdef:
        cnp.int_t node_id, i, n = len(base_level_nodes), m = len(closed_nodes)
        pair[cnp.int_t, cnp.float_t] node_pair

    for i in range(n):
        # NB: for node_i in open_boundary raises a compiling error with nogil.
        node_id = base_level_nodes[i]
        node_pair = pair[cnp.int_t, cnp.float_t](node_id, z[node_id])
        to_do.push(node_pair)
        receivers[node_id] = node_id
        outlet_nodes[node_id] = node_id
        done[node_id] = 1; done_n_ptr[0] +=1

    for i in range(m):
        node_id = closed_nodes[i]
        receivers[node_id] = node_id
        outlet_nodes[node_id] = node_id
        done[node_id] = 1; done_n_ptr[0] +=1
        # NB: done_n_ptr[0]: method to dereference the pointer. The
        # cython.operator.dereference doesn't seem to work on this pointer to a
        # long-type variable.

# 4. Functions necessary for the flow direction processing (Steps #11 - 20)
###########################################################################

@cython.boundscheck(False)
cdef void _set_flooded_and_outlet(cnp.int_t donor_id, cnp.float_t [:] z,
    cnp.int_t [:] receivers, cnp.int_t [:] outlet_nodes,
    cnp.int_t [:] depression_outlet_nodes, cnp.int_t [:] flooded_nodes,
    cnp.float_t [:] depression_depths, cnp.float_t [:] depression_free_elevations,
    cnp.int_t flooded_status, cnp.int_t bad_index,
    cnp.float_t min_elevation_relative_diff) nogil:
    """ Updates the base-level outlet nodes (outlet_nodes), the depression outlet
    nodes (depression_outlet_nodes), the flooded status (flooded_nodes), and the
    depths of the depressions (depression_depths) for the node donor_id depending
    on its surface z value and the one of its receiver outlet and neighbors.

    Parameters
    ----------
    donor_id: long.
        Id of the node giving the flow.
    z: memoryview(double).
        Values of the surface where the flow is directed, for each node.
    receivers: memoryview(long).
        Ids of the receiver nodes for each node considered as a donor, ordered by
        donor id.
    outlet_nodes: memoryview(long).
        Ids of the base-level outlet node, for each donor node, ordered by the donor
        node ids.
    depression_outlet_nodes: memoryview(long).
        Outlet nodes of the depression.
    flooded_nodes: memoryview(bool).
        Flooded status for each node.
    depression_depths: memoryview(double).
        Depths of the depression (if existing) below the level of the depression outlet
        for each node, ordered by node id.
    depression_free_elevations: memoryview(double).
        Elevation of the surface corrected from depressions.
    flooded_status: long.
        Constant for flooded status.
    bad_index: long.
        Constant for bad index.
    min_elevation_relative_diff: double
        Minimum relative difference in elevation for the depression_free_elevations
        surface.
    """
    cdef:
        cnp.int_t receiver_id = receivers[donor_id]
        cnp.int_t receiver_depression_outlet = (
            receiver_id if (
                depression_outlet_nodes[receiver_id] == bad_index
            ) else depression_outlet_nodes[receiver_id]
        )
        cnp.int_t receiver_outlet = (
            receiver_id if (
                outlet_nodes[receiver_id] == bad_index
            ) else outlet_nodes[receiver_id]
        )
    outlet_nodes[donor_id] = receiver_outlet
    if z[donor_id] < z[receiver_depression_outlet]:
        depression_outlet_nodes[donor_id] = receiver_depression_outlet
        flooded_nodes[donor_id] = flooded_status
        depression_depths[donor_id] = z[receiver_depression_outlet] - z[donor_id]
        depression_free_elevations[donor_id] = (
            (1 + min_elevation_relative_diff) * depression_free_elevations[receiver_id]
        )

@cython.boundscheck(False)
cdef void _set_receiver(cnp.int_t donor_id, cnp.int_t receiver_id,
    cnp.int_t [:] receivers, cnp.int_t [:] done, cnp.int_t* done_n_ptr) nogil:
    """ Updates the receiver (receivers) and the process statuses (done) of the donor
    node donor_id.

    Parameters
    ----------
    donor_id: long
        Id of the donor node.
    receiver_id: long
        Id of the receiver node.
    receivers: memoryview(long)
        Ids of the receivers for all donor nodes, ordered by the id of the donor nodes.
    done: memoryview(bool)
        Process statuses for all nodes. 1 for done. 0 otherwise.
    """
    cdef cnp.int_t done_n

    receivers[donor_id] = receiver_id
    done[donor_id] = 1 # 1 For True
    done_n_ptr[0] += 1

@cython.boundscheck(False)
cdef void _set_donor_properties(cnp.int_t donor_id, cnp.int_t receiver_id,
    cnp.int_t [:] sorted_pseudo_tails, const cnp.int_t [:,:] head_start_end_indexes,
    const cnp.int_t [:] sorted_dupli_links,
    cnp.float_t [:] sorted_dupli_gradients, cnp.float_t [:] z,
    cnp.float_t [:] steepest_slopes, cnp.int_t [:] links_to_receivers) nogil:
    """ Updates the steepest_slopes and the links_to_receivers of a donor in function
    of the slopes with its neighbors and the head-tail links of the grid. Steepest
    slope is set to 0. if the donor is in a depression.

    Parameters
    ----------
    donor_id: long
        Id of the donor node.
    receiver_id: long.
        Id of the receiver node.
    sorted_pseudo_tails: memoryview(long).
        All tails and heads for all existing links in the grid, ordered by
        increasing head and tail ids in sorted_pseudo_heads (see FlowRouter.py file).
    head_start_end_indexes: memoryview(longxlong).
        Start (array[0, :] and end (array[1, :]) index of all possible value nodes in
        the sorted_pseudo_heads (ordered by increasing head and tail ids, see
        FlowRouter.py file).
    sorted_dupli_links: memoryview(long).
        All existing links in the grid, duplicated, ordered by
        increasing head and tail ids in sorted_pseudo_heads (see FlowRouter.py file).
    sorted_dupli_gradients: memoryview(double).
        Gradients of the surface for all links duplicated, ordered by  increasing
        head and tail ids in sorted_pseudo_heads.
    z: memoryview(double).
        Values of the surface where the flow is directed, for each node.
    steepest_slopes: memoryview(double).
        Steepest slope for each node.
    links_to_receivers: memoryview(long).
        Ids of the links between a donor and a receiver, ordered by the ids of the
        donor nodes.
    """

    # range of indexes where donor_id is founded in sorted_pseudo_heads
    cdef:
        cnp.int_t idx1 = head_start_end_indexes[0, donor_id]
        cnp.int_t idx2 = head_start_end_indexes[1, donor_id] + 1


        cnp.int_t [:] s = sorted_pseudo_tails[idx1:idx2]
        cnp.int_t n = len(s), c = -1, i

    # loop to bypass the impossibility to use
    # sorted_pseudo_tails[idx1:idx2] == receiver_id with memoryviews
    for i in range(n):
        if s[i] == receiver_id:
            c = idx1 + i
            break
    steepest_slopes[donor_id] = (
        sorted_dupli_gradients[c] if (
            z[receiver_id] <= z[donor_id]
        ) else 0
    )
    links_to_receivers[donor_id] = sorted_dupli_links[c]

#######################################################################################

# Main functions to direct flow
###############################

@cython.boundscheck(False)
cdef void _direct_flow_c(cnp.int_t nodes_n, const cnp.int_t[:] base_level_nodes,
    const cnp.int_t[:] closed_nodes, cnp.int_t[:] sorted_pseudo_tails,
    cnp.float_t[:] sorted_dupli_gradients, const cnp.int_t[:] sorted_dupli_links,
    const cnp.int_t[:, :] head_start_end_indexes, cnp.int_t [:] outlet_nodes,
    cnp.int_t [:] depression_outlet_nodes, cnp.int_t[:] flooded_nodes,
    cnp.float_t[:] depression_depths, cnp.float_t[:] depression_free_elevations,
    cnp.int_t[:] links_to_receivers, cnp.int_t[:] receivers,
    cnp.float_t[:] steepest_slopes, cnp.float_t[:] z, cnp.int_t flooded_status,
    cnp.int_t bad_index, cnp.int_t neighbors_max_number,
    cnp.float_t min_elevation_relative_diff):
    """
    Main function implementing the flow directing through breaching depressions.
    Updates outlet_nodes, depression_outlet_nodes, flooded_nodes, links_to_receivers,
    receivers, steepest_slopes.

    Params
    ------
    nodes_n: long.
        Number of nodes to handle.
    base_level_nodes: memoryview(long).
        Base-level nodes.
    closed_nodes: memoryview(long).
        Closed nodes.
    sorted_pseudo_tails: memoryview(long).
        All tails and heads for all existing links in the grid, ordered by
        increasing head and tail ids in sorted_pseudo_heads (see FlowRouter.py file).
    sorted_dupli_gradients: memoryview(double).
        Gradients of the surface for all links duplicated, ordered by increasing
        head and tail ids in sorted_pseudo_heads.
    sorted_dupli_links: memoryview(long).
        All existing links in the grid, duplicated, ordered by
        increasing head and tail ids in sorted_pseudo_heads (see FlowRouter.py file).
    head_start_end_indexes: memoryview(longxlong).
        Start (array[0, :] and end (array[1, :]) index of all possible value nodes in
        the sorted_pseudo_heads (ordered by increasing head and tail ids, see
        FlowRouter.py file).
    outlet_nodes: memoryview(long).
        Ids of the base-level outlet node, for each donor node, ordered by the donor
        node ids.
    depression_outlet_nodes: memoryview(long).
        Outlet nodes of the depression, ordered by the donor node ids.
    flooded_nodes: memoryview(bool).
        Flooded status for each node, ordered by node ids.
    depression_depths: memoryview(double).
        Depths of the depression (if existing) below the level of the depression outlet
        for each node, ordered by node ids.
    depression_free_elevations: memoryview(double).
        Elevation of the surface corrected from depressions.
    links_to_receivers: memoryview(long).
        Ids of the links between a donor and a receiver, ordered by the ids of the
        donor nodes.
    receivers: memoryview(long)
        Ids of the receivers for all donor nodes, ordered by the id of the donor nodes.
    steepest_slopes: memoryview(double).
        Steepest slope for each node, ordered by node id.
    z: memoryview(double).
        Values of the surface where the flow is directed, for each node, ordered by
        node id.
    flooded_status: long.
        Constant for flooded status.
    bad_index: long.
        Constant for bad index.
    neighbors_max_number: long
        Maximum number of neighbors in the grid.
    min_elevation_relative_diff: double
        Minimum relative difference in elevation for the depression_free_elevations
        surface.
    """
    cdef:
        cnp.int_t [:] done
        cnp.int_t [:] tmp_neighbors
        cnp.int_t [:] neighbors_to_do
        _priority_queue to_do = _priority_queue(_compare_second)
        cnp.int_t receiver_id, donor_id, n, i, j, done_n
        cnp.int_t [:] neighbors
        pair[cnp.int_t, cnp.float_t] node_pair

    done = np.full(nodes_n, 0, dtype=int)
    tmp_neighbors = np.full(neighbors_max_number, 0, dtype=int)
    neighbors_to_do = np.array([], dtype=int)
    done_n = 0 # done_n is input only for MULTITHREADING, a future evolution, and
               # is not checked in this function.

    _init_flow_direction_queues(base_level_nodes, closed_nodes, z, to_do,
        receivers, outlet_nodes, done, &done_n)

    for j in range(nodes_n):
        # a while loop is possible here, but prefer for loop, with future
        # multithreading evolution.
        if to_do.empty() == True:
            break
        receiver_id = to_do.top().first; to_do.pop(); done[0] = 1

        # Get the neighbors to handle.
        idx1 = head_start_end_indexes[0, receiver_id]
        idx2 = head_start_end_indexes[1, receiver_id] + 1

        # Handle each neighbor.
        for i in range(idx1, idx2):
            donor_id = sorted_pseudo_tails[i];
            if done[donor_id] == 1: continue # 0 for False.
            _set_receiver(donor_id, receiver_id, receivers, done, &done_n)
            _set_flooded_and_outlet(donor_id, z, receivers, outlet_nodes,
                depression_outlet_nodes, flooded_nodes, depression_depths,
                depression_free_elevations, flooded_status, bad_index,
                min_elevation_relative_diff)
            _set_donor_properties(donor_id, receiver_id, sorted_pseudo_tails,
                head_start_end_indexes, sorted_dupli_links, sorted_dupli_gradients,
                z, steepest_slopes, links_to_receivers)
            node_pair = pair[cnp.int_t, cnp.float_t](donor_id, z[donor_id])
            to_do.push(node_pair)

def _direct_flow(cnp.int_t nodes_n, const cnp.int_t[:] base_level_nodes,
    const cnp.int_t[:] closed_nodes, cnp.int_t[:] sorted_pseudo_tails,
    cnp.float_t[:] sorted_dupli_gradients, const cnp.int_t[:] sorted_dupli_links,
    const cnp.int_t[:, :] head_start_end_indexes, cnp.int_t [:] outlet_nodes,
    cnp.int_t [:] depression_outlet_nodes, cnp.int_t[:] flooded_nodes,
    cnp.float_t[:] depression_depths, cnp.float_t[:] depression_free_elevations,
    cnp.int_t[:] links_to_receivers, cnp.int_t[:] receivers,
    cnp.float_t[:] steepest_slopes, cnp.float_t[:] z, cnp.int_t flooded_status,
    cnp.int_t bad_index, cnp.int_t neighbors_max_number,
    cnp.float_t min_elevation_relative_diff):
    """
    Main function calling the function that implements flow directing through
    breaching depressions. Updates outlet_nodes, depression_outlet_nodes,
    flooded_nodes, links_to_receivers, receivers, steepest_slopes.

    Params
    ------
    nodes_n: long.
        Number of nodes to handle.
    base_level_nodes: memoryview(long).
        Base-level nodes.
    closed_nodes: memoryview(long).
        Closed nodes.
    sorted_pseudo_tails: memoryview(long).
        All tails and heads for all existing links in the grid, ordered by
        increasing head and tail ids in sorted_pseudo_heads (see FlowRouter.py file).
    sorted_dupli_gradients: memoryview(double).
        Gradients of the surface for all links duplicated, ordered by increasing
        head and tail ids in sorted_pseudo_heads.
    sorted_dupli_links: memoryview(long).
        All existing links in the grid, duplicated, ordered by
        increasing head and tail ids in sorted_pseudo_heads (see FlowRouter.py file).
    head_start_end_indexes: memoryview(longxlong).
        Start (array[0, :] and end (array[1, :]) index of all possible value nodes in
        the sorted_pseudo_heads (ordered by increasing head and tail ids, see
        FlowRouter.py file).
    outlet_nodes: memoryview(long).
        Ids of the base-level outlet node, for each donor node, ordered by the donor
        node ids.
    depression_outlet_nodes: memoryview(long).
        Outlet nodes of the depression, ordered by the donor node ids.
    flooded_nodes: memoryview(bool).
        Flooded status for each node, ordered by node ids.
    depression_depths: memoryview(double).
        Depths of the depression (if existing) below the level of the depression outlet
        for each node, ordered by node ids.
    depression_free_elevations: memoryview(double).
        Elevation of the surface corrected from depressions.
    links_to_receivers: memoryview(long).
        Ids of the links between a donor and a receiver, ordered by the ids of the
        donor nodes.
    receivers: memoryview(long)
        Ids of the receivers for all donor nodes, ordered by the id of the donor nodes.
    steepest_slopes: memoryview(double).
        Steepest slope for each node, ordered by node id.
    z: memoryview(double).
        Values of the surface where the flow is directed, for each node, ordered by
        node id.
    flooded_status: long.
        Constant for flooded status.
    bad_index: long.
        Constant for bad index.
    neighbors_max_number: long
        Maximum number of neighbors in the grid.
    min_elevation_relative_diff: double
        Minimum relative difference in elevation for the depression_free_elevations
        surface.
    """
    _direct_flow_c(nodes_n, base_level_nodes, closed_nodes,
                    sorted_pseudo_tails, sorted_dupli_gradients,
                    sorted_dupli_links, head_start_end_indexes,
                    outlet_nodes, depression_outlet_nodes,
                    flooded_nodes, depression_depths, depression_free_elevations,
                    links_to_receivers, receivers, steepest_slopes, z,
                    flooded_status, bad_index, neighbors_max_number,
                    min_elevation_relative_diff)



#!/usr/bin/env python3
"""
Created on Fri May  1 14:02:18 2020

@author: gtucker

Adjusted on Thu Jan 19 2023
@author: bcampforts
"""

import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.components import ExponentialWeatherer


def test_create_weatherer_and_change_rate():
    grid = RasterModelGrid((3, 3), 1.0)
    grid.add_zeros("soil__depth", at="node")
    ew = ExponentialWeatherer(grid, soil_production_maximum_rate=0.0001)
    ew.maximum_weathering_rate = 0.0004
    np.testing.assert_array_equal(
        ew.maximum_weathering_rate, np.broadcast_to(0.0004, grid.number_of_nodes)
    )
    with pytest.raises(ValueError):
        ExponentialWeatherer(grid, soil_production_maximum_rate=-0.0001)
    with pytest.raises(ValueError):
        ExponentialWeatherer(grid, soil_production_decay_depth=-0.0001)


def test_run_weatherer():
    grid = RasterModelGrid((5, 5), 1.0)
    grid.add_ones("soil__depth", at="node")
    sp_max = np.random.rand(25)
    ew = ExponentialWeatherer(grid, soil_production_maximum_rate=sp_max)
    ew.run_one_step()

    new_value = np.random.rand(25)
    new_value[7] = -0.1
    with pytest.raises(ValueError):
        ew.maximum_weathering_rate = new_value
    with pytest.raises(ValueError):
        ew.decay_depth = new_value

    ew.maximum_weathering_rate = 0.0004
    np.testing.assert_array_equal(
        ew.maximum_weathering_rate, np.broadcast_to(0.0004, grid.number_of_nodes)
    )
    ew.decay_depth = 0.5
    np.testing.assert_array_equal(
        ew.decay_depth, np.broadcast_to(0.5, grid.number_of_nodes)
    )

    # Test analytical value
    ew.run_one_step()
    soil_production_rate = 0.0004 * np.exp(-grid.at_node["soil__depth"] / 0.5)
    grid.at_node["soil_production__rate"]
    np.testing.assert_array_equal(
        soil_production_rate[grid.core_nodes],
        grid.at_node["soil_production__rate"][grid.core_nodes],
    )



import pytest
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.io.netcdf import from_netcdf, to_netcdf


@pytest.mark.parametrize("include", ((), [], set(), None))
def test_include_keyword_is_empty(tmpdir, format, include):
    grid = RasterModelGrid((4, 3), xy_spacing=(2, 5), xy_of_lower_left=(-2.0, 10.0))
    grid.add_ones("elev", at="node")
    grid.add_zeros("elev", at="link")
    grid.add_empty("temp", at="node")

    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)
        actual = from_netcdf("test.nc", include=include)
        assert len(actual.at_node) == 0
        assert len(actual.at_link) == 0


@pytest.mark.parametrize("include", ("*", ("*",), ("at_node:*", "at_link:*")))
@pytest.mark.parametrize("exclude", (None, ()))
def test_include_everything(tmpdir, format, include, exclude):
    grid = RasterModelGrid((4, 3), xy_spacing=(2, 5), xy_of_lower_left=(-2.0, 10.0))
    grid.add_ones("elev", at="node")
    grid.add_zeros("elev", at="link")
    grid.add_empty("temp", at="node")

    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)
        actual = from_netcdf("test.nc", include=include)
        assert set(actual.at_node) == {"elev", "temp"}
        assert set(actual.at_link) == {"elev"}


@pytest.mark.parametrize(
    "include,exclude", [(("*", "*")), ((None, None)), (([], None))]
)
def test_exclude_everything(tmpdir, format, include, exclude):
    grid = RasterModelGrid((4, 3), xy_spacing=(2, 5), xy_of_lower_left=(-2.0, 10.0))
    grid.add_ones("elev", at="node")
    grid.add_zeros("elev", at="link")
    grid.add_empty("temp", at="node")

    with tmpdir.as_cwd():
        to_netcdf(grid, "test.nc", format=format)
        actual = from_netcdf("test.nc", include=include, exclude=exclude)
        assert len(actual.at_node) == 0
        assert len(actual.at_link) == 0


@pytest.mark.parametrize(
    "grid_type", ["HexModelGrid", "RadialModelGrid", "RasterModelGrid"]
)
def test_from_grid(datadir, grid_type):
    grid = from_netcdf(datadir / f"test-{grid_type}.nc")
    assert grid.__class__.__name__ == grid_type
    assert_array_equal(grid.at_node["elev"], 1.0)
    assert_array_equal(grid.at_node["temp"], 1.0)
    assert_array_equal(grid.at_link["elev"], 0.0)



import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.data_record import DataRecord

grid = RasterModelGrid((3, 3))


@pytest.fixture
def dr_time():
    time = [0.0]
    data_vars = {"mean_elevation": (["time"], np.array([100]))}
    attrs = {"time_units": "y"}
    return DataRecord(grid=grid, time=time, data_vars=data_vars, attrs=attrs)


@pytest.fixture
def dr_item():
    my_items2 = {
        "grid_element": np.array(("node", "link"), dtype=str),
        "element_id": np.array([1, 3]),
    }
    return DataRecord(grid=grid, items=my_items2)


@pytest.fixture
def dr_2dim():
    time = [0.0]
    my_items3 = {
        "grid_element": np.array([["node"], ["link"]]),
        "element_id": np.array([[1], [3]]),
    }
    my_data_vars = {
        "mean_elevation": (["time"], [110.0]),
        "item_size": (["item_id", "time"], np.array([[0.3], [0.4]])),
    }
    return DataRecord(grid=grid, time=time, items=my_items3, data_vars=my_data_vars)


@pytest.fixture
def dr_nodim():
    return DataRecord(grid=grid)



import pytest

from landlab.components import NetworkSedimentTransporter


def test_basic_init(example_nmg, example_parcels, example_flow_director):
    _ = NetworkSedimentTransporter(
        example_nmg,
        example_parcels,
        example_flow_director,
        bed_porosity=0.3,
        g=9.81,
        fluid_density=1000,
        transport_method="WilcockCrowe",
    )


def test_bad_flow_director(example_nmg, example_parcels):
    with pytest.raises(ValueError):
        NetworkSedimentTransporter(
            example_nmg,
            example_parcels,
            "bad_fd",
            bed_porosity=0.03,
            g=9.81,
            fluid_density=1000,
            transport_method="WilcockCrowe",
        )


def test_bad_network_model_grid(example_parcels, example_flow_director):
    with pytest.raises(ValueError):
        NetworkSedimentTransporter(
            "bad_nmg",
            example_parcels,
            example_flow_director,
            bed_porosity=0.03,
            g=9.81,
            fluid_density=1000,
            transport_method="WilcockCrowe",
        )


def test_bad_parcels(example_nmg, example_flow_director):
    with pytest.raises(ValueError):
        NetworkSedimentTransporter(
            example_nmg,
            "bad_parcels",
            example_flow_director,
            bed_porosity=0.03,
            g=9.81,
            fluid_density=1000,
            transport_method="WilcockCrowe",
        )


def test_bad_porosity(example_nmg, example_parcels, example_flow_director):
    with pytest.raises(ValueError):
        NetworkSedimentTransporter(
            example_nmg,
            example_parcels,
            example_flow_director,
            bed_porosity=-0.03,
            g=9.81,
            fluid_density=1000,
            transport_method="WilcockCrowe",
        )


def test_bad_transport_method(example_nmg, example_parcels, example_flow_director):
    with pytest.raises(ValueError):
        NetworkSedimentTransporter(
            example_nmg,
            example_parcels,
            example_flow_director,
            bed_porosity=0.03,
            g=9.81,
            fluid_density=1000,
            transport_method="bad_transport_method",
        )



from functools import partial
from itertools import tee

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib.patches import Patch
from scipy.interpolate import interp1d


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def plot_layers(
    elevation_at_layer,
    x=None,
    sea_level=0.0,
    color_water=(0.8, 1.0, 1.0),
    color_bedrock=(0.8, 0.8, 0.8),
    color_layer=None,
    layer_line_width=0.5,
    layer_line_color="k",
    title=None,
    x_label="Distance",
    y_label="Elevation",
    legend_location="lower left",
):
    """Plot a stack of sediment layers as a cross section.

    Create a plot of the elevation sediment layers, including surfaces for
    sea level and bedrock.

    Parameters
    ----------
    elevation_at_layer : array-like of shape *(n_layers, n_stacks)*
        Elevation to each layer along the profile. Layers are provided
        row-by-row, with the bottom-most layer being the first row.
    x : array-like, optional
        Distance to each stack along the cross-section. If not provided,
        stack number will be used.
    sea_level : float, optional
        Elevation of sea level.
    color_water : tuple of float, optional
        Tuple of *(red, green, blue)* values for water.
    color_bedrock : tuple of float, optional
        Tuple of *(red, green, blue)* values for bedrock.
    color_layer : string, optional
        Colormap to use to color in the layers.
    layer_line_width : float, optional
        Width of line used to plot layer surfaces.
    layer_line_color : string, optional
        Color of the line used to plot layer surfaces.
    title : string, optional
        Text to be used for the graph's title. The default is to not
        include a title.
    x_label : string, optional
        Text to be used for the x (horizontal) axis label.
    y_label : string, optional
        Text to be used for the y (vertical) axis label.
    legend_location : string, optional
        Where to put the legend.
    """
    elevation_at_layer = np.asarray(elevation_at_layer)
    elevation_at_layer = np.expand_dims(
        elevation_at_layer,
        axis=tuple(np.arange(2 - elevation_at_layer.ndim)),
    )

    if len(elevation_at_layer) == 0:
        raise ValueError(
            f"no layers to plot (elevation_at_layer.shape is {np.shape(elevation_at_layer)}"
        )

    if x is None:
        x = np.arange(elevation_at_layer.shape[1])

    top_surface = elevation_at_layer[-1]
    bottom_surface = elevation_at_layer[0]

    if len(elevation_at_layer) > 0:
        _plot_layers(
            x,
            elevation_at_layer,  # [layers_to_plot],
            color=color_layer,
            lc=layer_line_color,
            lw=layer_line_width,
        )
    _plot_water(x, top_surface, sea_level=sea_level, fc=color_water)
    _plot_bedrock(x, bottom_surface, fc=color_bedrock)
    _plot_surface(x, top_surface, sea_level=sea_level)

    legend_location and _plot_legend(
        legend_location, color_water=color_water, color_bedrock=color_bedrock
    )

    plt.xlabel(x_label)
    plt.ylabel(y_label)
    title and plt.title(title)

    plt.show()


def _plot_water(x, y, sea_level=0.0, fc=(0.8, 1.0, 1.0)):
    x_water, y_water = _insert_shorelines(x, y, sea_level=sea_level)
    if fc is not None:
        plt.fill_between(x_water, y_water, sea_level, where=y_water <= sea_level, fc=fc)

    water_surface = np.full_like(x_water, sea_level, dtype=float)
    water_surface[y_water > sea_level] = np.nan
    plt.plot(x_water, water_surface, color="b")


def _plot_bedrock(x, y, fc=(0.8, 0.8, 0.8)):
    if fc is not None:
        plt.fill_between(
            x,
            y,
            np.full_like(y, y.min()),
            color=fc,
        )
    plt.plot(x, y, color="k")


def _plot_surface(x, y, sea_level=0.0):
    under_water = y <= sea_level
    plt.plot(x[~under_water], y[~under_water], color="g")
    plt.plot(x[under_water], y[under_water], color="b")


def _plot_layers(x, layers, color=None, lc="k", lw=0.5):
    if color is not None:
        cmap = cm.get_cmap(color)
        for layer, (lower, upper) in enumerate(pairwise(layers)):
            plt.fill_between(
                x,
                lower,
                upper,
                fc=cmap(layer * 256 // len(layers)),
            )
    plt.plot(
        x,
        layers.T,
        color=lc,
        linewidth=lw,
    )


def _plot_legend(legend_location, color_water=None, color_bedrock=None):
    legend_item = partial(Patch, edgecolor="k", linewidth=0.5)
    items = [
        ("Ocean", color_water),
        ("Bedrock", color_bedrock),
    ]
    legend = [legend_item(label=label, fc=color) for label, color in items if color]
    legend and plt.legend(handles=legend, loc=legend_location)


def _insert_shorelines(x, y, sea_level=0.0):
    """Insert shorelines into x-y arrays.

    Examples
    --------
    >>> from landlab.plot.layers import _insert_shorelines
    >>> _insert_shorelines([0, 1, 2],[2, 1, -1])
    (array([ 0. ,  1. ,  1.5,  2. ]), array([ 2.,  1.,  0., -1.]))
    """
    x, y = np.asarray(x, dtype=float), np.asarray(y, dtype=float)

    y_relative_to_sea_level = y - sea_level
    shorelines = _search_zero_crossings(y_relative_to_sea_level)
    x_of_shoreline = _interp_zero_crossings(x, y_relative_to_sea_level, shorelines)

    return (
        np.insert(x, shorelines + 1, x_of_shoreline),
        np.insert(y, shorelines + 1, sea_level),
    )


def _search_zero_crossings(y):
    """Search an array for changes in sign between elements.

    Parameters
    ----------
    y : array-like
        Input array to check for sign changes.

    Returns
    -------
    int
        Indices into *y* where a sign has changed.

    Examples
    --------
    >>> from landlab.plot.layers import _search_zero_crossings

    The returned index is to the element before the zero-crossing.

    >>> list(_search_zero_crossings([2, 1, -1]))
    [1]
    >>> list(_search_zero_crossings([-2, -2, 1, 2]))
    [1]
    >>> list(_search_zero_crossings([-2, -2, 1, 2, -1]))
    [1, 3]

    These are not zero-crossings.

    >>> len(_search_zero_crossings([2, 0, 0, -2])) == 0
    True
    >>> len(_search_zero_crossings([2, 0, 1])) == 0
    True
    >>> len(_search_zero_crossings([2, 3, 4])) == 0
    True
    >>> len(_search_zero_crossings([0, 0, 0])) == 0
    True
    """
    sign = np.sign(y)

    # zeros = sign == 0
    # if not np.all(zeros):
    #     while np.any(zeros):
    #         sign[zeros] = np.roll(sign, 1)[zeros]
    #         zeros = sign == 0

    # return np.where(sign[1:] != sign[:-1])[0]
    return np.where(sign[1:] * sign[:-1] < 0)[0]


def _interp_zero_crossings(x, y, shorelines):
    """Interpolate between adjacent elements to find x-locations of zero-crossings.

    Parameters
    ----------
    x : array-like
        Distances.
    y : array-like
        Elevations.
    shorelines : array-like of int
        Indices to shoreline elements.

    Returns
    -------
    array of float
        Distances to interpolated shorelines.

    Examples
    --------
    >>> from landlab.plot.layers import _interp_zero_crossings
    >>> _interp_zero_crossings([0, 1, 2], [1, -1, -1], [0])
    array([ 0.5])
    >>> _interp_zero_crossings([0, 1, 2, 3], [1, -1, -1, 4], [0, 2])
    array([ 0.5, 2.2])
    """
    x_of_shoreline = []
    for shoreline in shorelines:
        coast = slice(shoreline, shoreline + 2)

        # for scipy<1.10 interp1d requires x and y to have at least two elements,
        # which is not the case if theshoreline is the last element.
        x_of_shoreline.append(
            interp1d(np.broadcast_to(y[coast], 2), np.broadcast_to(x[coast], 2))(0.0)
        )

    return np.asarray(x_of_shoreline)



"""The Basic Modeling Interface."""


class Error(Exception):
    """Base class for BMI exceptions."""

    pass


class FatalError(Exception):
    """Raise this exception if an unrecoverable error was found."""

    pass


class BadVarNameError(Error):
    """Exception to indicate a bad input/output variable name."""

    def __init__(self, name):
        super().__init__()
        self.name = name

    def __str__(self):
        return self.name


class MissingModelAttributeError(Error):
    """Raise this exception if a component is missing a required attribute."""

    def __init__(self, attrib):
        super().__init__()
        self.attrib = attrib

    def __str__(self):
        return self.attrib


class TimeBoundsError(Error):
    """Raise this exception if a component updates beyond its time horizon."""

    pass


class BmiGridType(int):
    """Base type to indicate the type of a BMI model's grid.

    :code: Grid type code as an int
    :name: Name of the grid type as a string
    """

    def __new__(cls, code, name):
        obj = super().__new__(cls, code)
        obj.name = name
        return obj

    def __str__(self):
        return self.name

    def __repr__(self):
        return "BmiGridType(%d, %s)" % (self, self.name)


GRID_TYPE_UNKNOWN = BmiGridType(-1, "Unknown")
GRID_TYPE_NONE = BmiGridType(0, "No grid")
GRID_TYPE_UNIFORM = BmiGridType(1, "Uniform rectilinear")
GRID_TYPE_RECTILINEAR = BmiGridType(2, "Rectilinear")
GRID_TYPE_STRUCTURED = BmiGridType(3, "Structured")
GRID_TYPE_UNSTRUCTURED = BmiGridType(4, "Unstructured")


class BmiBase:
    """Definition of the Basic Modeling Interface."""

    def initialize(self, file_name):
        """Initialize model.

        :file_name: String of configuration file
        """
        pass

    def update(self, **kwds):
        """Update model by one time step."""
        pass

    def finalize(self):
        """Clean-up model."""
        pass

    def get_input_var_names(self):
        """Get names of input variables to the model as standard names.

        :returns: A list of input standard names as strings
        """
        pass

    def get_output_var_names(self):
        """Get names of output variables to the model as standard names.

        :returns: A list of output standard names as strings
        """
        pass

    def get_var_type(self, var_name):
        """Get type of an exchange item."""
        pass

    def get_var_units(self, var_name):
        """Get units of an exchange item."""
        pass

    def get_var_rank(self, var_name):
        """Rank of exchange item."""
        pass

    def get_time_step(self):
        """Model time step."""
        pass

    def get_start_time(self):
        """Model start time."""
        pass

    def get_current_time(self):
        """Current time of model."""
        pass

    def get_end_time(self):
        """Model stop time."""
        pass


class BmiExtendedBase:
    """An extension interface for a BMI."""

    def update_until(self, time):
        """Update model until some time.

        :time: Update duration
        """
        pass

    def run_model(self):
        """Initialize, run, and finalize a model."""
        pass


class BmiUnstructured:
    """BMI for a model that uses an unstructured grid."""

    def get_x(self, name):
        """Get x-coordinates of grid nodes."""
        pass

    def get_y(self, name):
        """Get y-coordinates of grid nodes."""
        pass

    def get_connectivity(self, name):
        """Get cell connectivity."""
        pass

    def get_offset(self, name):
        """Get cell offset."""
        pass


class BmiStructured:
    """BMI for a model that uses a structured grid."""

    def get_grid_shape(self, name):
        """Get shape of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_x(self, name):
        """Get x-coordinates of grid nodes."""
        pass

    def get_y(self, name):
        """Get y-coordinates of grid nodes."""
        pass


class BmiRectilinear:
    """BMI for a model that uses a rectilinear grid."""

    def get_grid_shape(self, name):
        """Get shape of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_columns(self, name):
        """Get coordinates of grid columns."""
        pass

    def get_rows(self, name):
        """Get coordinates of grid rows."""
        pass


class BmiUniformRectilinear:
    """BMI for a model that exposes a uniform rectilinear grid."""

    def get_grid_shape(self, name):
        """Get shape of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_grid_spacing(self, name):
        """Get spacing of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_grid_origin(self, name):
        """Get origin of grid for variable, name.

        :name: Standard name
        """
        pass


class BmiNoGrid:
    """BMI for a model that does not have a grid."""



"""
Tests written by Jeff Keck and Allison Pfeiffer
"""

import numpy as np
import pytest

from landlab.components import (
    BedParcelInitializerArea,
    BedParcelInitializerDepth,
    BedParcelInitializerDischarge,
    BedParcelInitializerUserD50,
)
from landlab.components.network_sediment_transporter.bed_parcel_initializers import (
    _determine_approx_parcel_volume,
    calc_d50_dArea_scaling,
    calc_d50_depth,
    calc_d50_discharge,
    calc_total_parcel_volume,
)
from landlab.data_record import DataRecord


# Basic test: that you can call the four initializers, create DataRecord
def test_call_area_BPI(example_nmg2):
    initialize_parcels = BedParcelInitializerArea(
        example_nmg2, drainage_area_coefficient=0.18, drainage_area_exponent=-0.12
    )
    parcels = initialize_parcels()
    assert isinstance(parcels, DataRecord)


def test_call_discharge_BPI(example_nmg2):
    initialize_parcels = BedParcelInitializerDischarge(
        example_nmg2, discharge_at_link=np.full(example_nmg2.number_of_links, 10.0)
    )
    parcels = initialize_parcels()
    assert isinstance(parcels, DataRecord)


def test_call_depth_BPI(example_nmg2):
    depth_at_link = np.full(example_nmg2.number_of_links, 10.0)  # m
    initialize_parcels = BedParcelInitializerDepth(
        example_nmg2,
        flow_depth_at_link=depth_at_link,
    )
    parcels = initialize_parcels()
    assert isinstance(parcels, DataRecord)


def test_call_userD50_BPI(example_nmg2):
    d50_at_link = np.full(example_nmg2.number_of_links, 0.08)  # m
    initialize_parcels = BedParcelInitializerUserD50(
        example_nmg2,
        user_d50=d50_at_link,
    )
    parcels = initialize_parcels()
    assert isinstance(parcels, DataRecord)


# %% Tests of BPI base class helper functions


def test_calc_total_parcel_volume():
    vole = 12.5
    vol = calc_total_parcel_volume(5, 5, 0.5)
    np.testing.assert_allclose(vol, vole, rtol=1e-4)


def test_det_approx_parcel_volume():
    total_parcel_volume_at_link = 10
    median_number_of_starting_parcels = 50
    pvole = 0.2
    pvol = _determine_approx_parcel_volume(
        total_parcel_volume_at_link, median_number_of_starting_parcels
    )
    np.testing.assert_allclose(pvol, pvole, rtol=1e-4)


# %% Test to make sure we're getting correct correct d50 correct_values


def test_calc_D50_discharge():
    """test calc D50 grain size give correct values"""
    width = 20
    slope = 0.01
    discharge = 100
    mannings_n = 0.05
    gravity = 9.81
    rho_water = 1000
    rho_sediment = 3000
    tau_c_50 = 0.05

    D50e = 0.173286

    D50 = calc_d50_discharge(
        width,
        slope,
        discharge,
        mannings_n,
        gravity,
        rho_water,
        rho_sediment,
        tau_c_50,
    )

    np.testing.assert_allclose(D50, D50e, rtol=1e-4)


class Test_calc_d50_dArea_scaling:
    def test_normal_d50_dArea_scaling(self):
        """test normal values 1"""
        D50e = 0.13182567
        a = 0.1
        n = 0.12
        CA = 10  # km^2
        D50 = calc_d50_dArea_scaling(drainage_area=CA, a=a, n=n)
        np.testing.assert_allclose(D50, D50e, rtol=1e-4)

    def test_normal_d50_dArea_scaling_2(self):
        """test normal values 2"""
        D50e = 0.1
        a = 0.1
        n = 0
        CA = 10  # km^2
        D50 = calc_d50_dArea_scaling(drainage_area=CA, a=a, n=n)
        np.testing.assert_allclose(D50, D50e, rtol=1e-4)


def test_calc_d50_depth():
    """test normal values for calc_d50_depth"""
    D50e = 0.1
    D50 = calc_d50_depth(
        slope=0.01,
        flow_depth=1,
        tau_c_multiplier=1,
        rho_water=1000.0,
        rho_sediment=3000.0,
        tau_c_50=0.05,
    )
    np.testing.assert_allclose(D50, D50e, rtol=1e-4)


# %% Test for errors raised
def test_D50_not_specified(example_nmg2):
    """test ValueError exception raised if D50 input is None"""
    with pytest.raises(ValueError):
        initialize_parcels = BedParcelInitializerUserD50(
            example_nmg2,
            user_d50=None,
        )
        initialize_parcels()


# %% Test for expected correct values


class Test_BedParcelInitializer:
    def test_normal_BPI(self, example_nmg2):
        """
        Minimum attributes specified, most attributes should use
        defaults specified at instantiation. uses BedParcelInitializerDischarge.
        """
        np.random.seed(seed=5)
        discharge = np.full(example_nmg2.number_of_links, 1.0)
        initialize_parcels = BedParcelInitializerDischarge(
            example_nmg2,
            discharge_at_link=discharge,
            median_number_of_starting_parcels=1,
        )

        parcels = initialize_parcels()

        # create expected values and get values from datarecord
        GEe = np.expand_dims(["link"] * 8, axis=1).astype(object)
        GE = parcels.dataset["grid_element"]
        EIe = np.expand_dims(np.array([0, 1, 2, 3, 4, 5, 5, 6]), axis=1)
        EI = parcels.dataset["element_id"]
        SLe = np.array([0, 1, 2, 3, 4, 5, 5, 6])
        SL = parcels.dataset["starting_link"]
        ARe = np.ones(8) * 0
        AR = parcels.dataset["abrasion_rate"]
        De = np.ones(8) * 2650
        D = parcels.dataset["density"]
        TAe = np.expand_dims(np.zeros(8), axis=1)
        #   TAe = np.array([[ 0.08074127], [ 0.7384403 ], [ 0.44130922], [ 0.15830987],
        #                 [ 0.87993703], [ 0.27408646], [ 0.41423502], [ 0.29607993]])
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.expand_dims(np.ones(8), axis=1)
        AL = parcels.dataset["active_layer"]
        LL = parcels.dataset["location_in_link"]
        De = np.array(
            [
                [0.06802885],
                [0.06232028],
                [0.37674903],
                [0.06607135],
                [0.07391346],
                [0.29280262],
                [0.04609956],
                [0.05135768],
            ]
        )
        D = parcels.dataset["D"]
        V = parcels.dataset["volume"]

        assert list(GE.values) == list(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(D, De, rtol=1e-4)
        np.testing.assert_array_less(TAe, TA)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        np.testing.assert_array_less(LL, 1)
        np.testing.assert_array_less(0, D)
        np.testing.assert_array_less(0, V)

    def test_normal_BPI_abrasion(self, example_nmg2):
        """
        Test normal outputs of bed parcel initializer with abrasion.
        uses BedParcelInitializerDischarge.
        """
        np.random.seed(seed=5)
        discharge = np.full(example_nmg2.number_of_links, 1.0)
        initialize_parcels = BedParcelInitializerDischarge(
            example_nmg2,
            discharge_at_link=discharge,
            median_number_of_starting_parcels=1,
            abrasion_rate=0.1,
        )

        parcels = initialize_parcels()

        # create expected values and get values from datarecord
        GEe = np.expand_dims(["link"] * 8, axis=1).astype(object)
        GE = parcels.dataset["grid_element"]
        EIe = np.expand_dims(np.array([0, 1, 2, 3, 4, 5, 5, 6]), axis=1)
        EI = parcels.dataset["element_id"]
        SLe = np.array([0, 1, 2, 3, 4, 5, 5, 6])
        SL = parcels.dataset["starting_link"]
        ARe = np.ones(8) * 0.1
        AR = parcels.dataset["abrasion_rate"]
        De = np.ones(8) * 2650
        D = parcels.dataset["density"]
        # TAe = np.array([[ 0.08074127], [ 0.7384403 ], [ 0.44130922], [ 0.15830987],
        #                 [ 0.87993703], [ 0.27408646], [ 0.41423502], [ 0.29607993]])
        TAe = np.expand_dims(np.zeros(8), axis=1)
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.expand_dims(np.ones(8), axis=1)
        AL = parcels.dataset["active_layer"]
        De = np.array(
            [
                [0.06802885],
                [0.06232028],
                [0.37674903],
                [0.06607135],
                [0.07391346],
                [0.29280262],
                [0.04609956],
                [0.05135768],
            ]
        )
        D = parcels.dataset["D"]
        V = parcels.dataset["volume"]

        assert list(GE.values) == list(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(D, De, rtol=1e-4)
        # np.testing.assert_allclose(TA, TAe, rtol = 1e-4)
        np.testing.assert_array_less(TAe, TA)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        # np.testing.assert_allclose(LL, LLe, rtol = 1e-4)
        np.testing.assert_array_less(0, D)
        np.testing.assert_array_less(0, V)



#! /usr/bin/env python
"""Unit tests for landlab.components.threshold_eroder.py"""
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from landlab import HexModelGrid, RasterModelGrid
from landlab.components import (
    FlowAccumulator,
    FlowDirectorSteepest,
    PriorityFloodFlowRouter,
    ThresholdEroder,
)

try:
    PriorityFloodFlowRouter.load_richdem()
except ModuleNotFoundError:
    with_richdem = False
else:
    with_richdem = True


@pytest.mark.skipif(not with_richdem, reason="richdem is not installed")
def test_topography_rasterGrid():
    # %%
    mg = RasterModelGrid((5, 5))
    mg.set_closed_boundaries_at_grid_edges(False, False, False, False)
    z = np.array(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 1.0, 10.0, 1.0, 0.0]
        + [0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )
    _ = mg.add_field("topographic__elevation", z, at="node")

    # Instantiate Flow director (steepest slope type) and Threshold Eroder
    fdir = PriorityFloodFlowRouter(mg)
    th_ero = ThresholdEroder(mg, slope_crit=0.6)

    # Run the components for ten short timepsteps
    for _t in range(2):
        fdir.run_one_step()
        th_ero.run_one_step()

    # Check final topography

    assert_array_almost_equal(
        mg.at_node["topographic__elevation"],
        np.array(
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.6, 0.6, 0.0, 0.0, 0.6, 1.2, 0.6]
            + [0.0, 0.0, 0.6, 0.6, 0.6, 0.0, 0.0, 0.0, 0.0, 0.0, 0]
        ),
        verbose=True,
    )


@pytest.mark.skipif(not with_richdem, reason="richdem is not installed")
def test_topo_soil_rasterGrid():
    # %%
    mg = RasterModelGrid((5, 5))
    mg.set_closed_boundaries_at_grid_edges(False, False, False, False)
    z = np.array(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0]
        + [0.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )
    topo = mg.add_zeros("topographic__elevation", at="node")
    bed = mg.add_field("bedrock__elevation", z, at="node")
    soil = mg.add_ones("soil__depth", at="node")
    soil[mg.boundary_nodes] = 0
    topo[:] = soil + bed

    # Instantiate Flow director (steepest slope type) and Threshold Eroder
    fdir = PriorityFloodFlowRouter(mg)
    th_ero = ThresholdEroder(mg, slope_crit=0.6)

    # Run the components for ten short timepsteps

    for _t in range(2):
        fdir.run_one_step()
        th_ero.run_one_step()

    # Check final topography
    assert_array_almost_equal(
        mg.at_node["bedrock__elevation"],
        np.array(
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.6, 0.6, 0.0, 0.0, 0.6, 1, 0.6]
            + [0.0, 0.0, 0.6, 0.6, 0.6, 0.0, 0.0, 0.0, 0.0, 0.0, 0]
        ),
        verbose=True,
    )
    assert_array_almost_equal(
        mg.at_node["soil__depth"],
        np.array(
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        ),
        verbose=True,
    )


# %%
def test_topography_hexGrid():
    # %%
    hmg = HexModelGrid((8, 8))

    topo = hmg.add_zeros("topographic__elevation", at="node")
    topo[hmg.core_nodes] += 100

    # Instantiate Flow director (steepest slope type) and Threshold Eroder
    fdir = FlowDirectorSteepest(hmg)
    fa = FlowAccumulator(hmg)
    th_ero = ThresholdEroder(hmg, slope_crit=0.6)

    # Run the components for ten short timepsteps
    for _t in range(5):
        fdir.run_one_step()
        fa.run_one_step()
        th_ero.run_one_step()
    assert_array_almost_equal(
        hmg.at_node["topographic__elevation"],
        np.array(
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.6, 0.6, 0.6, 0.6]
            + [0.6, 0.6, 0.0, 0.0, 0.6, 1.2, 1.2, 1.2, 1.2, 1.2, 1.2, 0.6, 0.0, 0.0]
            + [0.6, 1.2, 1.8, 1.8, 1.8, 1.8, 1.8, 1.2, 0.6, 0.0, 0.0, 0.6, 1.2, 1.8]
            + [1.8, 1.8, 1.8, 1.8, 1.8, 1.2, 0.6, 0.0, 0.0, 0.6, 1.2, 1.2, 1.2, 1.2]
            + [1.2, 1.2, 1.2, 0.6, 0.0, 0.0, 0.6, 0.6, 0.6, 0.6, 0.6, 0.6, 0.6, 0.6]
            + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        ),
    )



import numpy as np
import pandas as pd
import pytest

from landlab.components.network_sediment_transporter.sediment_pulser_at_links import (
    SedimentPulserAtLinks,
)
from landlab.components.network_sediment_transporter.sediment_pulser_base import (
    SedimentPulserBase,
)
from landlab.components.network_sediment_transporter.sediment_pulser_each_parcel import (
    SedimentPulserEachParcel,
)


def always_time_to_pulse(time):
    return True


def time_to_pulse_list(time):
    Ptime = [19, 20, 22, 23, 24, 75, 76]
    return time in Ptime


def test_call_SedimentPulserBase(example_nmg2):
    """test exception raised if SedimentPulserBase is called"""
    grid = example_nmg2
    make_pulse = SedimentPulserBase(grid)

    with pytest.raises(NotImplementedError) as exc_info:
        _ = make_pulse()
    msg_e = "the base component has no call method"
    assert exc_info.match(msg_e)


# @pytest.mark.xfail(reason = "TDD, test class is not yet implemented")
class Test_SedimentPulserAtLinks:
    def test_normal_1(self, example_nmg2):
        """only time specified, links and number parcels specified,
        should use defaults in base class"""
        grid = example_nmg2
        time_to_pulse = always_time_to_pulse

        make_pulse = SedimentPulserAtLinks(grid, time_to_pulse=time_to_pulse)

        time = 11
        links = [0]
        n_parcels_at_link = [10]
        parcels = make_pulse(
            time=time, links=links, n_parcels_at_link=n_parcels_at_link
        )

        # create expected values and get values from datarecord
        GEe = np.expand_dims(["link"] * 10, axis=1)
        GE = parcels.dataset["grid_element"]
        EIe = np.expand_dims(np.zeros(10).astype(int), axis=1)
        EI = parcels.dataset["element_id"]
        SLe = np.zeros(10).astype(int)
        SL = parcels.dataset["starting_link"]
        ARe = np.ones(10) * 0
        AR = parcels.dataset["abrasion_rate"]
        roe = np.ones(10) * 2650
        ro = parcels.dataset["density"]
        TAe = np.expand_dims(np.ones(10) * time, axis=1)
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.expand_dims(np.ones(10), axis=1)
        AL = parcels.dataset["active_layer"]
        LL = parcels.dataset["location_in_link"]
        D50 = 0.05  # default grain size parameters
        D84_D50 = 2.1
        D_sd = D50 * D84_D50 - D50
        D = parcels.dataset["D"]
        Ve = np.expand_dims(np.ones(10) * 0.5, axis=1)
        V = parcels.dataset["volume"]

        # test
        assert list(GE.values) == list(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(ro, roe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        assert (LL < 0).any() >= 0 and (LL < 0).any() <= 1  # must be between 0 and 1
        # check mean randomly selected grain size with within 3 standard deviations
        # of specified mean
        assert D.mean() < (D50 + 3 * D_sd) and D.mean() > (D50 - 3 * D_sd)
        np.testing.assert_allclose(V, Ve, rtol=1e-4)

    def test_normal_2(self, example_nmg2):
        """only D50 specified, all other parcel attributes should
        use defaults in base class"""

        grid = example_nmg2
        time_to_pulse = always_time_to_pulse

        make_pulse = SedimentPulserAtLinks(grid, time_to_pulse=time_to_pulse)
        time = 11
        links = [2, 6]
        n_parcels_at_link = [2, 3]
        D50 = [0.3, 0.12]
        parcels = make_pulse(
            time=time, links=links, n_parcels_at_link=n_parcels_at_link, D50=D50
        )

        # create expected values and get values from datarecord
        GEe = np.expand_dims(["link"] * 5, axis=1)
        GE = parcels.dataset["grid_element"]
        EIe = np.expand_dims(np.array([2, 2, 6, 6, 6]), axis=1)
        EI = parcels.dataset["element_id"]
        SLe = np.array([2, 2, 6, 6, 6])
        SL = parcels.dataset["starting_link"]
        ARe = np.ones(5) * 0
        AR = parcels.dataset["abrasion_rate"]
        roe = np.ones(5) * 2650
        ro = parcels.dataset["density"]
        TAe = np.expand_dims(np.ones(5) * time, axis=1)
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.expand_dims(np.ones(5), axis=1)
        AL = parcels.dataset["active_layer"]
        LL = parcels.dataset["location_in_link"]
        D50_1 = D50[0]  # grain size
        D84_D50_1 = 2.1
        D_sd_1 = D50_1 * D84_D50_1 - D50_1

        D50_2 = D50[1]  # grain size
        D84_D50_2 = 2.1
        D_sd_2 = D50_2 * D84_D50_2 - D50_2

        D = parcels.dataset["D"]
        Ve = np.expand_dims(np.ones(5) * 0.5, axis=1)
        V = parcels.dataset["volume"]

        assert list(GE.values) == list(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(ro, roe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        assert (LL < 0).any() >= 0 and (LL < 0).any() <= 1  # must be between 0 and 1
        # check mean randomly selected grain size with within 3 standard deviations
        # of specified mean
        assert D[0:2].mean() < (D50_1 + 3 * D_sd_1) and D[0:2].mean() > (
            D50_1 - 3 * D_sd_1
        )
        assert D[2:].mean() < (D50_2 + 3 * D_sd_2) and D[2:].mean() > (
            D50_2 - 3 * D_sd_2
        )
        np.testing.assert_allclose(V, Ve, rtol=1e-4)

    def test_normal_3(self, example_nmg2):
        """two pulses. First, only time, links and number of parcels specified,
        uses defaults in base class for all other parcel attributes
        second, two parcels in link two and three parcels in link six
        are added and all attributes specified"""

        grid = example_nmg2
        time_to_pulse = always_time_to_pulse

        make_pulse = SedimentPulserAtLinks(grid, time_to_pulse=time_to_pulse)

        time = 11
        links = [0]
        n_parcels_at_link = [2]
        parcels = make_pulse(
            time=time, links=links, n_parcels_at_link=n_parcels_at_link
        )

        time = 12
        links = [2, 6]
        n_parcels_at_link = [2, 3]
        D50 = [0.3, 0.12]
        D84_D50 = [2.1, 1.5]
        parcel_volume = [1, 0.5]
        rho_sediment = [2650, 2500]
        abrasion_rate = [0.1, 0.3]
        parcels = make_pulse(
            time=time,
            links=links,
            n_parcels_at_link=n_parcels_at_link,
            D50=D50,
            D84_D50=D84_D50,
            parcel_volume=parcel_volume,
            rho_sediment=rho_sediment,
            abrasion_rate=abrasion_rate,
        )

        # create expected values and get values from datarecord
        GEe = [
            ["link", np.nan],
            ["link", np.nan],
            [np.nan, "link"],
            [np.nan, "link"],
            [np.nan, "link"],
            [np.nan, "link"],
            [np.nan, "link"],
        ]
        GE = parcels.dataset["grid_element"]
        EIe = np.array(
            [
                [0.0, np.nan],
                [0.0, np.nan],
                [np.nan, 2.0],
                [np.nan, 2.0],
                [np.nan, 6.0],
                [np.nan, 6.0],
                [np.nan, 6.0],
            ]
        )
        EI = parcels.dataset["element_id"]
        SLe = np.array([0.0, 0.0, 2.0, 2.0, 6.0, 6.0, 6.0])
        SL = parcels.dataset["starting_link"]
        ARe = np.array([0.0, 0.0, 0.1, 0.1, 0.3, 0.3, 0.3])
        AR = parcels.dataset["abrasion_rate"]
        roe = np.array([2650.0, 2650.0, 2650.0, 2650.0, 2500.0, 2500.0, 2500.0])
        ro = parcels.dataset["density"]
        TAe = np.array(
            [
                [11.0, np.nan],
                [11.0, np.nan],
                [np.nan, 12.0],
                [np.nan, 12.0],
                [np.nan, 12.0],
                [np.nan, 12.0],
                [np.nan, 12.0],
            ]
        )
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.array(
            [
                [1.0, np.nan],
                [1.0, np.nan],
                [np.nan, 1.0],
                [np.nan, 1.0],
                [np.nan, 1.0],
                [np.nan, 1.0],
                [np.nan, 1.0],
            ]
        )
        AL = parcels.dataset["active_layer"]
        LL = parcels.dataset["location_in_link"].values
        LL = LL[~np.isnan(LL)]
        D = parcels.dataset["D"].values
        D[~np.isnan(D)]
        Ve = np.array(
            [
                [0.5, np.nan],
                [0.5, np.nan],
                [np.nan, 1.0],
                [np.nan, 1.0],
                [np.nan, 0.5],
                [np.nan, 0.5],
                [np.nan, 0.5],
            ]
        )
        V = parcels.dataset["volume"]

        # test
        assert str(GE.values.tolist()) == str(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(ro, roe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        assert (LL < 0).any() >= 0 and (LL < 0).any() <= 1  # must be between 0 and 1
        # grain size must be greater than 0
        assert (D > 0).any()
        np.testing.assert_allclose(V, Ve, rtol=1e-4)

    def test_special_1(self, example_nmg2):
        """user entered time is not a pulse time, calling instance returns
        the original parcels datarecord, which is None if there is no
        original datarecord"""

        grid = example_nmg2
        time_to_pulse = time_to_pulse_list
        # np.random.seed(seed=5)

        make_pulse = SedimentPulserAtLinks(grid, time_to_pulse=time_to_pulse)

        time = 11
        links = [0]
        n_parcels_at_link = [2]
        parcels = make_pulse(
            time=time, links=links, n_parcels_at_link=n_parcels_at_link
        )

        assert parcels is None


# @pytest.mark.xfail(reason = "TDD, test class is not yet implemented")
class Test_SedimentPulserEachParcel:
    def test_normal_1(self, example_nmg2):
        """minimum attributes specified in Pulse, attributes should use
        defaults specified at instantiation"""

        grid = example_nmg2
        # np.random.seed(seed=5)

        make_pulse = SedimentPulserEachParcel(grid)

        PulseDF = pd.DataFrame(
            {
                "pulse_volume": [0.2, 1, 1.1, 0.5],
                "link_#": [1, 3, 5, 2],
                "normalized_downstream_distance": [0.8, 0.7, 0.5, 0.2],
            }
        )
        time = 7
        parcels = make_pulse(time, PulseDF)

        # create expected values and get values from datarecord
        GEe = np.expand_dims(["link"] * 7, axis=1)
        GE = parcels.dataset["grid_element"]
        EIe = np.expand_dims(np.array([1, 3, 3, 5, 5, 5, 2]), axis=1)
        EI = parcels.dataset["element_id"]
        SLe = np.array([1, 3, 3, 5, 5, 5, 2])
        SL = parcels.dataset["starting_link"]
        ARe = np.ones(7) * 0
        AR = parcels.dataset["abrasion_rate"]
        roe = np.ones(7) * 2650
        ro = parcels.dataset["density"]
        TAe = np.expand_dims(np.ones(7) * time, axis=1)
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.expand_dims(np.ones(7), axis=1)
        AL = parcels.dataset["active_layer"]
        LLe = np.array([[0.8], [0.7], [0.7], [0.5], [0.5], [0.5], [0.2]])
        LL = parcels.dataset["location_in_link"]
        D50 = 0.05  # default grain size parameters
        D84_D50 = 2.1
        D_sd = D50 * D84_D50 - D50
        D = parcels.dataset["D"]
        Ve = np.array([[0.2], [0.5], [0.5], [0.5], [0.5], [0.1], [0.5]])
        V = parcels.dataset["volume"]

        assert list(GE.values) == list(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(ro, roe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        np.testing.assert_allclose(LL, LLe, rtol=1e-4)
        assert D.mean() < (D50 + 3 * D_sd) and D.mean() > (D50 - 3 * D_sd)
        np.testing.assert_allclose(V, Ve, rtol=1e-4)

    def test_normal_2(self, example_nmg2):
        """all attributes specified in Pulse"""

        grid = example_nmg2
        # np.random.seed(seed=5)

        make_pulse = SedimentPulserEachParcel(grid)

        PulseDF = pd.DataFrame(
            {
                "pulse_volume": [0.2, 1, 1.1, 0.5],
                "link_#": [1, 3, 5, 2],
                "normalized_downstream_distance": [0.8, 0.7, 0.5, 0.2],
                "D50": [0.15, 0.2, 0.22, 0.1],
                "D84_D50": [1, 1, 1, 1],
                "abrasion_rate": [0.01, 0.02, 0.005, 0.03],
                "rho_sediment": [2650, 2300, 2750, 2100],
                "parcel_volume": [0.1, 1, 1, 0.2],
            }
        )
        time = 7
        parcels = make_pulse(time, PulseDF)

        # create expected values and get values from datarecord
        GEe = np.expand_dims(["link"] * 8, axis=1)
        GE = parcels.dataset["grid_element"]
        EIe = np.expand_dims(np.array([1, 1, 3, 5, 5, 2, 2, 2]), axis=1)
        EI = parcels.dataset["element_id"]
        SLe = np.array([1, 1, 3, 5, 5, 2, 2, 2])
        SL = parcels.dataset["starting_link"]
        ARe = np.array([0.01, 0.01, 0.02, 0.005, 0.005, 0.03, 0.03, 0.03])
        AR = parcels.dataset["abrasion_rate"]
        roe = np.array([2650, 2650, 2300, 2750, 2750, 2100, 2100, 2100])
        ro = parcels.dataset["density"]
        TAe = np.expand_dims(np.ones(8) * time, axis=1)
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.expand_dims(np.ones(8), axis=1)
        AL = parcels.dataset["active_layer"]
        LLe = np.array([[0.8], [0.8], [0.7], [0.5], [0.5], [0.2], [0.2], [0.2]])
        LL = parcels.dataset["location_in_link"]
        De = np.array([[0.15], [0.15], [0.2], [0.22], [0.22], [0.1], [0.1], [0.1]])
        D = parcels.dataset["D"]
        Ve = np.array([[0.1], [0.1], [1], [1], [0.1], [0.2], [0.2], [0.1]])
        V = parcels.dataset["volume"]

        assert list(GE.values) == list(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(ro, roe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        np.testing.assert_allclose(LL, LLe, rtol=1e-4)
        np.testing.assert_allclose(D, De, rtol=1e-4)
        np.testing.assert_allclose(V, Ve, rtol=1e-4)

    def test_normal_3(self, example_nmg2):
        """two pulses. First, only minimum attributes specified,
        second, two parcels in link two and three parcels in link six
        are added and all attributes specified"""

        grid = example_nmg2
        np.random.seed(seed=5)

        make_pulse = SedimentPulserEachParcel(grid)

        PulseDF = pd.DataFrame(
            {
                "pulse_volume": [0.2, 1],
                "link_#": [1, 3],
                "normalized_downstream_distance": [0.8, 0.7],
            }
        )
        time = 7
        parcels = make_pulse(time, PulseDF)

        PulseDF = pd.DataFrame(
            {
                "pulse_volume": [1.1, 0.5],
                "link_#": [5, 2],
                "normalized_downstream_distance": [0.5, 0.2],
                "D50": [0.22, 0.1],
                "D84_D50": [2.1, 1.5],
                "abrasion_rate": [0.005, 0.03],
                "density": [2750, 2100],
                "parcel_volume": [1, 0.2],
            }
        )
        time = 8
        parcels = make_pulse(time, PulseDF)

        # create expected values and get values from datarecord
        GEe = [
            ["link", np.nan],
            ["link", np.nan],
            ["link", np.nan],
            [np.nan, "link"],
            [np.nan, "link"],
            [np.nan, "link"],
            [np.nan, "link"],
            [np.nan, "link"],
        ]
        GE = parcels.dataset["grid_element"]
        EIe = np.array(
            [
                [1.0, np.nan],
                [3.0, np.nan],
                [3.0, np.nan],
                [np.nan, 5.0],
                [np.nan, 5.0],
                [np.nan, 2.0],
                [np.nan, 2.0],
                [np.nan, 2.0],
            ]
        )
        EI = parcels.dataset["element_id"]
        SLe = np.array([1, 3, 3, 5, 5, 2, 2, 2])
        SL = parcels.dataset["starting_link"]
        ARe = np.array([0.0, 0.0, 0.0, 0.005, 0.005, 0.03, 0.03, 0.03])
        AR = parcels.dataset["abrasion_rate"]
        roe = np.array([2650.0, 2650.0, 2650.0, 2650.0, 2650.0, 2650.0, 2650.0, 2650.0])
        ro = parcels.dataset["density"]
        TAe = np.array(
            [
                [7.0, np.nan],
                [7.0, np.nan],
                [7.0, np.nan],
                [np.nan, 8.0],
                [np.nan, 8.0],
                [np.nan, 8.0],
                [np.nan, 8.0],
                [np.nan, 8.0],
            ]
        )
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.array(
            [
                [1.0, np.nan],
                [1.0, np.nan],
                [1.0, np.nan],
                [np.nan, 1.0],
                [np.nan, 1.0],
                [np.nan, 1.0],
                [np.nan, 1.0],
                [np.nan, 1.0],
            ]
        )
        AL = parcels.dataset["active_layer"]
        LLe = np.array(
            [
                [0.8, np.nan],
                [0.7, np.nan],
                [0.7, np.nan],
                [np.nan, 0.5],
                [np.nan, 0.5],
                [np.nan, 0.2],
                [np.nan, 0.2],
                [np.nan, 0.2],
            ]
        )
        LL = parcels.dataset["location_in_link"]
        D = parcels.dataset["D"].values
        D[~np.isnan(D)]
        Ve = np.array(
            [
                [0.2, np.nan],
                [0.5, np.nan],
                [0.5, np.nan],
                [np.nan, 1.0],
                [np.nan, 0.1],
                [np.nan, 0.2],
                [np.nan, 0.2],
                [np.nan, 0.1],
            ]
        )
        V = parcels.dataset["volume"]

        # test
        assert str(GE.values.tolist()) == str(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(ro, roe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        np.testing.assert_allclose(LL, LLe, rtol=1e-4)
        # grain size must be greater than 0
        assert (D > 0).any()
        np.testing.assert_allclose(V, Ve, rtol=1e-4)

    def test_normal_4(self, example_nmg2):
        """Series of pulses using both the SedimentPulserAtlinks and
        SedimentPulserEachParcel."""

        grid = example_nmg2
        np.random.seed(seed=5)  # use the random seed for this test

        # define the initial parcels datarecord
        make_pulse_links = SedimentPulserAtLinks(
            grid, time_to_pulse=always_time_to_pulse
        )

        # pulse 1
        time = 7
        links = [0]
        n_parcels_at_link = [2]
        parcels = make_pulse_links(
            time=time, links=links, n_parcels_at_link=n_parcels_at_link
        )

        # datarecord is input to SedimentPulserEachParcel, now both pulser
        # instances will point to the same datarecord when called
        make_pulse_pulseDF = SedimentPulserEachParcel(parcels=parcels, grid=grid)

        # pulse 2
        PulseDF = pd.DataFrame(
            {
                "pulse_volume": [1.1],
                "link_#": [5],
                "normalized_downstream_distance": [0.5],
                "D50": [0.22],
                "D84_D50": [1],
                "abrasion_rate": [0.005],
                "density": [2750],
                "parcel_volume": [1],
            }
        )
        time = 8
        parcels = make_pulse_pulseDF(time, PulseDF)

        # pulse 3
        time = 9
        links = [5]
        n_parcels_at_link = [1]
        parcels = make_pulse_links(
            time=time, links=links, n_parcels_at_link=n_parcels_at_link
        )

        # pulse 4
        PulseDF = pd.DataFrame(
            {
                "pulse_volume": [0.5],
                "link_#": [6],
                "normalized_downstream_distance": [0.2],
            }
        )
        time = 10
        parcels = make_pulse_pulseDF(time, PulseDF)

        # create expected values and get values from datarecord
        GEe = [
            ["link", np.nan, np.nan, np.nan],
            ["link", np.nan, np.nan, np.nan],
            [np.nan, "link", np.nan, np.nan],
            [np.nan, "link", np.nan, np.nan],
            [np.nan, np.nan, "link", np.nan],
            [np.nan, np.nan, np.nan, "link"],
        ]
        GE = parcels.dataset["grid_element"]
        EIe = np.array(
            [
                [0.0, np.nan, np.nan, np.nan],
                [0.0, np.nan, np.nan, np.nan],
                [np.nan, 5.0, np.nan, np.nan],
                [np.nan, 5.0, np.nan, np.nan],
                [np.nan, np.nan, 5.0, np.nan],
                [np.nan, np.nan, np.nan, 6.0],
            ]
        )
        EI = parcels.dataset["element_id"]
        SLe = np.array([0, 0, 5, 5, 5, 6])
        SL = parcels.dataset["starting_link"]
        ARe = np.array([0.0, 0.0, 0.005, 0.005, 0.0, 0.0])
        AR = parcels.dataset["abrasion_rate"]
        D = parcels.dataset["density"]
        TAe = np.array(
            [
                [7.0, np.nan, np.nan, np.nan],
                [7.0, np.nan, np.nan, np.nan],
                [np.nan, 8.0, np.nan, np.nan],
                [np.nan, 8.0, np.nan, np.nan],
                [np.nan, np.nan, 9.0, np.nan],
                [np.nan, np.nan, np.nan, 10.0],
            ]
        )
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.array(
            [
                [1.0, np.nan, np.nan, np.nan],
                [1.0, np.nan, np.nan, np.nan],
                [np.nan, 1.0, np.nan, np.nan],
                [np.nan, 1.0, np.nan, np.nan],
                [np.nan, np.nan, 1.0, np.nan],
                [np.nan, np.nan, np.nan, 1.0],
            ]
        )
        AL = parcels.dataset["active_layer"]
        LLe = np.array(
            [
                [0.20671916, np.nan, np.nan, np.nan],
                [0.91861091, np.nan, np.nan, np.nan],
                [np.nan, 0.5, np.nan, np.nan],
                [np.nan, 0.5, np.nan, np.nan],
                [np.nan, np.nan, 0.2968005, np.nan],
                [np.nan, np.nan, np.nan, 0.2],
            ]
        )
        LL = parcels.dataset["location_in_link"]
        D = parcels.dataset["D"].values
        D[~np.isnan(D)]
        Ve = np.array(
            [
                [0.5, np.nan, np.nan, np.nan],
                [0.5, np.nan, np.nan, np.nan],
                [np.nan, 1.0, np.nan, np.nan],
                [np.nan, 0.1, np.nan, np.nan],
                [np.nan, np.nan, 0.5, np.nan],
                [np.nan, np.nan, np.nan, 0.5],
            ]
        )
        V = parcels.dataset["volume"]

        # test
        assert str(GE.values.tolist()) == str(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        np.testing.assert_allclose(LL, LLe, rtol=1e-4)
        # grain size must be greater than 0
        assert (D > 0).any()
        np.testing.assert_allclose(V, Ve, rtol=1e-4)

    def test_bad_1(self, example_nmg2):
        """test exception raised if instance is called without specifying the
        pulserDF"""

        grid = example_nmg2
        make_pulse = SedimentPulserEachParcel(grid)
        PulseDF = None
        time = 7

        with pytest.raises(ValueError) as exc_info:
            _ = make_pulse(time, PulseDF)
        msg_e = "PulseDF was not specified"
        assert exc_info.match(msg_e)

    def test_special_1(self, example_nmg2):
        """test that calling with an empty PulseDF returns the existing
        datarecord"""

        grid = example_nmg2
        np.random.seed(seed=5)

        make_pulse = SedimentPulserEachParcel(grid)

        PulseDF = pd.DataFrame(
            {
                "pulse_volume": [0.2, 1, 1.1, 0.5],
                "link_#": [1, 3, 5, 2],
                "normalized_downstream_distance": [0.8, 0.7, 0.5, 0.2],
            }
        )
        # call SedimentPulserEachParcel using setup in normal_test_1
        time1 = 7
        parcels = make_pulse(time1, PulseDF)

        # call again, using an empty PulserDF
        time2 = 8
        parcels = make_pulse(time2, pd.DataFrame([]))

        # create expected values and get values from datarecord
        GEe = np.expand_dims(["link"] * 7, axis=1)
        GE = parcels.dataset["grid_element"]
        EIe = np.expand_dims(np.array([1, 3, 3, 5, 5, 5, 2]), axis=1)
        EI = parcels.dataset["element_id"]
        SLe = np.array([1, 3, 3, 5, 5, 5, 2])
        SL = parcels.dataset["starting_link"]
        ARe = np.ones(7) * 0
        AR = parcels.dataset["abrasion_rate"]
        roe = np.ones(7) * 2650
        ro = parcels.dataset["density"]
        TAe = np.expand_dims(np.ones(7) * time1, axis=1)
        TA = parcels.dataset["time_arrival_in_link"]
        ALe = np.expand_dims(np.ones(7), axis=1)
        AL = parcels.dataset["active_layer"]
        LLe = np.array([[0.8], [0.7], [0.7], [0.5], [0.5], [0.5], [0.2]])
        LL = parcels.dataset["location_in_link"]
        De = np.array(
            [
                [0.06936526],
                [0.03911625],
                [0.30353682],
                [0.04147067],
                [0.05423609],
                [0.16176179],
                [0.02546817],
            ]
        )
        D = parcels.dataset["D"]
        Ve = np.array([[0.2], [0.5], [0.5], [0.5], [0.5], [0.1], [0.5]])
        V = parcels.dataset["volume"]

        assert list(GE.values) == list(GEe)
        np.testing.assert_allclose(EI, EIe, rtol=1e-4)
        np.testing.assert_allclose(SL, SLe, rtol=1e-4)
        np.testing.assert_allclose(AR, ARe, rtol=1e-4)
        np.testing.assert_allclose(ro, roe, rtol=1e-4)
        np.testing.assert_allclose(TA, TAe, rtol=1e-4)
        np.testing.assert_allclose(AL, ALe, rtol=1e-4)
        np.testing.assert_allclose(LL, LLe, rtol=1e-4)
        np.testing.assert_allclose(D, De, rtol=1e-4)
        np.testing.assert_allclose(V, Ve, rtol=1e-4)
